from meshiphi.dataloaders.dataloader_interface import DataLoaderInterface
from abc import abstractmethod

from pyproj import Transformer, CRS

import logging
import numpy as np
import xarray as xr
import pandas as pd
from rasterio.enums import Resampling

from meshiphi.utils import round_to_sigfig


class VectorDataLoader(DataLoaderInterface):
    '''
    Abstract class for all vector Datasets.
    '''
    def __init__(self, bounds, params):
        '''
        This is where large-scale operations are performed, 
        such as importing data, downsampling, reprojecting, and renaming 
        variables
        
        Args:
            bounds (Boundary): 
                Initial mesh boundary to limit scope of data ingest
            params (dict): 
                Values needed by dataloader to initialise. Unique to each
                dataloader

        Attributes:
            self.data (pd.DataFrame or xr.Dataset): 
                Data stored by dataloader to use when called upon by the mesh.
                Must be saved in mercator projection (EPSG:4326), with 
                coordinates names 'lat', 'long', and 'time' (if applicable).
            self.data_name (str): 
                Name of scalar variable. Must be the column name if self.data
                is pd.DataFrame. Must be variable if self.data is xr.Dataset
        '''
        # Translates parameters from config input to desired inputs
        params = self.add_default_params(params)
        logging.info(f"Initialising {params['dataloader_name']} dataloader")
        # Creates a class attribute for all keys in params
        for key, val in params.items():
            setattr(self, key, val)
            
        self.data = self.import_data(bounds)
        # Read in and manipulate data to standard form
        if 'files' in params:
            logging.info('\tFiles read:')
            for file in self.files:
                logging.info(f'\t\t{file}')
        # If need to downsample data
        self.data = self.downsample()
        # If need to reproject data
        if self.in_proj != self.out_proj:
            self.data = self.reproject(
                                in_proj  = self.in_proj,
                                out_proj = self.out_proj,
                                x_col    = self.x_col,
                                y_col    = self.y_col
                                )

        # Get data name from column name if not set in params
        if self.data_name is None:
            logging.debug('\tSetting self.data_name from column name')
            self.data_name = self.get_data_col_name()
        # or if set in params, set col name to data name
        else:
            logging.debug(f'\tSetting data column name to {self.data_name}')
            self.data = self.set_data_col_name(self.data_name.split(','))
        # Store data names in a list for easier access in future
        self.data_name_list = self.data_name.split(',')
        
        # Add magnitude and direction to dataset
        self.data = self.add_mag_dir()
        
        # Calculate fraction of boundary that data covers
        data_coverage = self.calculate_coverage(bounds)
        logging.info("\tMercator data range (roughly) covers "+\
                    f"{np.round(data_coverage*100,0).astype(int)}% "+\
                     "of initial boundary")
        # If there's 0 datapoints in the initial boundary, raise ValueError
        if data_coverage == 0:
            logging.error('\tDataloader has no data in initial region!')
            raise ValueError(f"Dataloader {params['dataloader_name']}"+\
                              " contains no data within initial region!")
        else:
            # Cut dataset down to initial boundary
            logging.info(
                "\tTrimming data to initial boundary: {min} to {max}".format(
                    min=(bounds.get_lat_min(), bounds.get_long_min()),
                    max=(bounds.get_lat_max(), bounds.get_long_max())
                ))
            
            self.data = self.trim_datapoints(bounds)

    @abstractmethod
    def import_data(self, bounds):
        '''
        User defined method for importing data from files, or even generating 
        data from scratch
                
        Returns:
            xr.Dataset or pd.DataFrame:
                Coordinates and data being imported from file \n
                if xr.Dataset, 
                    - Must have coordinates 'lat' and 'long'
                    - Should have multiple data variables
                    
                if pd.DataFrame, 
                    - Must have columns 'lat' and 'long'
                    - Should have multiple data columns
                    
                Downsampling and reprojecting happen in __init__() method
        '''
        pass
    
    def add_default_params(self, params):
        '''
        Set default values for all scalar dataloaders. This function should be
        overloaded to include any extra params for a specific dataloader
        
        Args:
            params (dict): 
                Dictionary containing attributes that are required for each 
                dataloader. 
            
        Returns:
            (dict): 
                Dictionary of attributes the dataloader will require, 
                completed with default values if not provided in config.
        '''
        # This should only pop up if using dataloader not in the factory
        if 'dataloader_name' not in params:
            params['dataloader_name'] = self.__class__.__name__
        
        if 'data_name' not in params:
            params['data_name'] = None

        if 'downsample_factors' not in params:
            params['downsample_factors'] = [1,1]

        if 'aggregate_type' not in params: 
            params['aggregate_type']  = 'MEAN'
            
        if 'min_dp' not in params:
            params['min_dp'] = 5
            
        if 'in_proj' not in params:
            params['in_proj'] = 'EPSG:4326'
            
        if 'out_proj' not in params:
            params['out_proj'] = 'EPSG:4326'
            
        if 'x_col' not in params:
            params['x_col'] = 'lat'

        if 'y_col' not in params:
            params['y_col'] = 'long'

        if 'fast_reprojection' not in params:
            params['fast_reprojection'] = False
            
        return params
    
    def add_mag_dir(self, data=None, data_names=None):
        '''
        Adds magnitude and direction variables/columns to data for easier
        retrieval of value 
        
        Args:
            data (pd.DataFrame or xr.Dataset):
                Data with 'lat' and 'long' columns/dimensions. Assumes that the
                existing data is in cartesian form (x and y components). 
                If None, will use self.data
            data_names (list):
                List of data columns/variables to use in calculation
                If None, will use self.data_name_list
                
        Returns:
            data (pd.DataFrame or xr.Dataset):
                Original dataset with two new columns/variables called 
                '_magnitude' and '_direction', containing the corresponding
                values for each.
        '''
        
        def add_mag_dir_to_df(data, names):
            '''
            Adds magnitude and direction columns to pd.DataFrame
            
            Args:
                Same as parent method
                
            Returns:
                Same as parent method
            '''
            x, y = names
            data['_magnitude'] = np.linalg.norm([data[x], data[y]], axis=0)
            data['_direction'] = np.arctan2(data[y], data[x])
            return data
        
        def add_mag_dir_to_xr(data, names):
            '''
            Adds magnitude and direction columns to xr.Dataset
            
            Args:
                Same as parent method
                
            Returns:
                Same as parent method
            '''
            x, y = names
            data['_magnitude'] = (data.dims, 
                                  np.linalg.norm([data[x].data, data[y].data], 
                                                 axis=0))
            data['_direction'] = (data.dims, 
                                  np.arctan2(data[y].data, data[x].data))
            return data
        
        # Set defaults if not passed to method
        if data is None:        data = self.data
        if data_names is None:  names = self.data_name_list
        
        # Perform operation on appropriate datatype
        if type(data) == pd.core.frame.DataFrame:
            return add_mag_dir_to_df(data, names)
        elif type(data) == xr.core.dataset.Dataset:
            return add_mag_dir_to_xr(data, names)

    def calculate_coverage(self, bounds, data=None):
        """
        Calculates percentage of boundary covered by dataset

        Args:
            bounds (Boundary): 
                Boundary being compared against
            data (pd.DataFrame or xr.Dataset): 
                Dataset with 'lat' and 'long' coordinates. 
                Extent calculated from min/max of these coordinates. 
                Defaults to objects internal dataset.
        
        Returns:
            float:
                Decimal fraction of boundary covered by the dataset
        """
        def calculate_coverage_from_df(bounds, data):
            data = data.dropna().reset_index()
            # If empty dataframe, 0% coverage
            if data.empty:
                return 0
            # If no valid coordinates within data range, 0% coverage
            elif data.lat.size == 0 or data.long.size == 0:
                return 0
            # Otherwise, calculate coverage, assuming rectangular region 
            # in mercator projection
            else:
                # Get range of latitude values
                data_lat_range = data.lat.max() - data.lat.min()
                bounds_lat_range = bounds.get_lat_max() - bounds.get_lat_min()
                # Get range of longitude values
                data_long_range = data.long.max() - data.long.min()
                bounds_long_range = bounds.get_long_max() - bounds.get_long_min()
                # Calcualte area of each region
                data_area = data_lat_range * data_long_range
                bounds_area = bounds_lat_range * bounds_long_range
                # If data area completely covers bounds, 100% coverage
                if data_area >= bounds_area:
                    return 1
                # Otherwise return decimal fraction
                else:
                    return data_area / bounds_area
                
                
        def calculate_coverage_from_xr(bounds, data):
            # Remove all NaN columns/rows
            data = data.dropna(dim="lat", how="all")
            data = data.dropna(dim="long", how="all")
            
            # If no valid coordinates within data range, 0% coverage
            if data.lat.size == 0 or data.long.size == 0:
                return 0
            # Otherwise, calculate coverage, assuming rectangular region 
            # in mercator projection
            else:
                # Get range of latitude values
                data_lat_range = data.lat.max().item() - data.lat.min().item()
                bounds_lat_range = bounds.get_lat_max() - bounds.get_lat_min()
                # Get range of longitude values
                data_long_range = data.long.max().item() - data.long.min().item()
                bounds_long_range = bounds.get_long_max() - bounds.get_long_min()
                # Calcualte area of each region
                data_area = data_lat_range * data_long_range
                bounds_area = bounds_lat_range * bounds_long_range
                # If data area completely covers bounds, 100% coverage
                if data_area >= bounds_area:
                    return 1
                # Otherwise return decimal fraction
                else:
                    return data_area / bounds_area
        # Use self.data if not no explicit dataset specified
        if data is None:
            data = self.data
        # Calculate data coverage fraction
        if type(self.data) == pd.core.frame.DataFrame:
            return calculate_coverage_from_df(bounds, data)
        elif type(self.data) == xr.core.dataset.Dataset:
            return calculate_coverage_from_xr(bounds, data)

    def trim_datapoints(self, bounds, data=None):
        '''
        Trims datapoints from self.data within boundary defined by 'bounds'.
        self.data can be pd.DataFrame or xr.Dataset
        
        Args:
            bounds (Boundary): Limits of lat/long/time to select data from

        Returns:
            pd.DataFrame or xr.Dataset: 
                Trimmed dataset in same format as self.data
        '''
        def trim_datapoints_from_df(data, bounds):
            '''
            Extracts data from a pd.DataFrame
            '''
            # Mask off any positions not within spatial bounds
            mask = (data['lat']  > bounds.get_lat_min())  & \
                   (data['lat']  <= bounds.get_lat_max())  & \
                   (data['long'] > bounds.get_long_min()) & \
                   (data['long'] <= bounds.get_long_max())
            # Mask with time if time column exists
            if 'time' in data.columns:
                mask &= (data['time'] >= bounds.get_time_min()) & \
                        (data['time'] <= bounds.get_time_max())
                        
            # Return column of data from within bounds
            return data.loc[mask]

        def trim_datapoints_from_xr(data, bounds):
            '''
            Extracts data from a xr.Dataset
            '''

            # Select data region within spatial bounds
            # NOTE slice in xarray is inclusive of bounds
            data = data.sel(lat=slice(bounds.get_lat_min(),  bounds.get_lat_max() ))
            data = data.sel(long=slice(bounds.get_long_min(), bounds.get_long_max()))
            # Select data region within temporal bounds if time exists as a coordinate
            if 'time' in data.coords.keys():
                data = data.sel(time=slice(bounds.get_time_min(),  bounds.get_time_max()))

            # Trim off any data on the min boundary to be consistent with df
            if bounds.get_lat_min() in data.lat:
                data = data.where(data.lat  != bounds.get_lat_min(), drop=True)
            if bounds.get_long_min() in data.long:
                data = data.where(data.long != bounds.get_long_min(), drop=True)
            
            # Return column of data from within bounds
            return data
        
        # If no specific data passed in, default to entire dataset
        if data is None:
            data = self.data
            
        # Skip trimming if data already completely within bounds
        if data.lat.min() >  bounds.get_lat_min() and \
           data.lat.max() <= bounds.get_lat_max() and \
           data.long.min() >  bounds.get_long_min() and \
           data.long.max() <= bounds.get_long_max():
            logging.debug('\tData is already trimmed to bounds!')
            return data
        
        if type(data) == pd.core.frame.DataFrame:
            return trim_datapoints_from_df(data, bounds)
        elif type(data) == xr.core.dataset.Dataset:
            return trim_datapoints_from_xr(data, bounds)

    def get_dp_from_coord(self, long=None, lat=None, return_coords=False):
        '''
        Extracts datapoint from self.data with lat and long specified in kwargs.
        self.data can be pd.DataFrame or xr.Dataset. Will return multiple values
        if one set of coordinates have multiple entries (e.g. time series data)
        
        Args:
            long (float): Longitude coordinate to search for
            lat (float) : Latitude coordinate to search for
            
        Returns:
            pd.Dataframe:  
                Column of data values with chosen lat/long. Could be many 
                datapoints because either bad data or multiple time steps 
        '''
        def get_dp_from_coord_df(data, names, long, lat, return_coords):
            '''
            Extracts data from a pd.DataFrame
            '''
            # Mask off any positions not within spatial bounds
            mask = (data['lat']  == lat)  & \
                   (data['long'] == long) 

            # Include lat/long/time if requested
            if return_coords: columns = list(data.columns)
            else:             columns = names
            # Return column of data from within bounds
            return data.loc[mask][columns]
        
        def get_dp_from_coord_xr(data, names, long, lat, return_coords):
            '''
            Extracts data from a xr.Dataset
            '''
            # Select data region within spatial bounds
            data = data.sel(lat=lat, long=long)
            # Cast as a pd.DataFrame
            data = data.to_dataframe().reset_index()
            # Include lat/long/time if requested
            if return_coords: columns = list(data.columns)
            else:             columns = names
            # Return column of data from within bounds
            return data[columns]
        
        # Ensure that lat and long provided
        assert (lat is not None) and (long) is not None, \
            'Must provide lat and long to this method!'
            
        # Choose which method to retrieve data based on input type
        if type(self.data) == pd.core.frame.DataFrame:
            return get_dp_from_coord_df(self.data, self.data_name_list, long, lat, return_coords)
        elif type(self.data) == xr.core.dataset.Dataset:
            return get_dp_from_coord_xr(self.data, self.data_name_list, long, lat, return_coords)
    
    def get_value(self, bounds, agg_type=None, skipna=True):
        '''
        Retrieve aggregated value from within bounds
        
        Args:
            aggregation_type (str): Method of aggregation of datapoints within
                bounds. Can be upper or lower case. 
                Accepts 'MIN', 'MAX', 'MEAN', 'MEDIAN', 'STD', 'COUNT'
            bounds (Boundary): Boundary object with limits of lat/long
            skipna (bool): Defines whether to propogate NaN's or not
                Default = True (ignore's NaN's)

        Returns:
            dict: 
                {variable (str): aggregated_value (float)}
                Aggregated value within bounds following aggregation_type
                
        Raises:
            ValueError: aggregation type not in list of available methods
        '''
        def get_value_from_df(dps, variable_names, bounds, agg_type, skipna):
            '''
            Aggregates a value from a pd.Series.
            
            Args:
                dps (pd.Series): Datapoints within boundary
                bounds (Boundary): 
                    Boundary dps was trimmed to. Not used for any calculations,
                    just the logging.debug message.
                agg_type (str):
                    Method of aggregation for the value, 
                    e.g. agg_type = 'MIN' => min(dps) returned 
                skipna (bool): 
                    Flag for whether NaN's should be included in aggregation. 
            
            Returns:
                np.float64: Aggregated value
            '''
            data_count = len(dps)
            logging.debug(f"\t{data_count} datapoints found for attribute '{self.data_name}' within bounds '{bounds}'")
            # If no data
            if data_count == 0:
                values = [np.nan, np.nan]
            # If want the number of datapoints
            elif agg_type =='COUNT':
                values = [data_count, data_count]
            elif agg_type == 'MIN':
                index = dps['_magnitude'].idxmin(skipna=skipna)
                values = [dps[name][index] for name in variable_names]
            elif agg_type == 'MAX':
                index = dps['_magnitude'].idxmax(skipna=skipna)
                values = [dps[name][index] for name in variable_names]
            elif agg_type == 'MEAN':
                values = [dps[name].mean(skipna=skipna) for name in variable_names]
            elif agg_type == 'STD':
                values = [dps[name].std(skipna=skipna) for name in variable_names]
            elif agg_type == 'MEDIAN':
                raise ValueError('Aggregation type "MEDIAN" is non-sensical for vector dataset!')
            else:
                raise ValueError(f'Unknown aggregation type {agg_type}')
            
            return values

        
        def get_value_from_xr(dps, variable_names, bounds, agg_type, skipna):
            '''
            Aggregates a value from a xr.DataArray.
            
            Args:
                dps (xr.DataArray): Datapoints within boundary
                bounds (Boundary): 
                    Boundary dps was trimmed to. Not used for any calculations,
                    just the logging.debug message.
                agg_type (str):
                    Method of aggregation for the value, 
                    e.g. agg_type = 'MIN' => min(dps) returned 
                skipna (bool): 
                    Flag for whether NaN's should be included in aggregation. 
            
            Returns:
                dict:
                    {variable_name: np.float64}: Aggregated value in a dictionary
            '''
            # Info on size of array
            data_count = dps._magnitude.size 
            logging.debug(f"\t{data_count} datapoints found for attribute '{self.data_name}' within bounds '{bounds}'")
            # If no data, return np.nan for each variable
            if data_count == 0:
                values = [np.nan for _ in variable_names]
            elif np.isnan(dps._magnitude).all():
                values = [np.nan for _ in variable_names]
            # If want count
            elif agg_type == 'COUNT':
                # If including nan's, just want size
                if skipna:  
                    values = [data_count for _ in variable_names]
                # Otherwise count non-nan values
                else:       
                    values = [dps[name].count().item() for name in variable_names]
            elif agg_type == 'MIN':
                # Get 2D index of minimum magnitude point
                index = np.unravel_index(
                            dps._magnitude.argmin(skipna=skipna), 
                            dps._magnitude.shape)
                # Get cartesian vector from index
                values = [dps[name][index].item() for name in variable_names]
            elif agg_type == 'MAX':
                # Get 2D index of minimum magnitude point
                index = np.unravel_index(
                            dps._magnitude.argmax(skipna=skipna), 
                            dps._magnitude.shape)
                # Get cartesian vector from index
                values = [dps[name][index].item() for name in variable_names]
            # And the rest self explanatory
            elif agg_type == 'MEAN':
                values = [dps[name].mean(skipna=skipna).item() for name in variable_names]
            elif agg_type == 'STD':
                values = [dps[name].std(skipna=skipna).item() for name in variable_names]
            elif agg_type == 'MEDIAN':
                raise ValueError('Aggregation type "MEDIAN" is non-sensical for vector dataset!')
            else:
                raise ValueError(f'Unknown aggregation type {agg_type}')
            
            return values
    

        # Set to params if no specific aggregate type specified
        if agg_type is None:
            agg_type = self.aggregate_type
            
        # Limit data to boundary
        dps = self.trim_datapoints(bounds)
        # Get list of values
        if type(self.data) == pd.core.frame.DataFrame:
            values = get_value_from_df(dps, self.data_name_list, bounds, agg_type, skipna)
        elif type(self.data) == xr.core.dataset.Dataset:
            values = get_value_from_xr(dps, self.data_name_list, bounds, agg_type, skipna)
            
        # Put in dict to map variable to values
        return {self.data_name_list[i]: values[i] for i in range(len(self.data_name_list))}

    def get_hom_condition(self, bounds, splitting_conds, agg_type='MEAN'):
        '''
        Retrieves homogeneity condition of data within boundary. 
         
        Args: 
            bounds (Boundary): Boundary object with limits of datarange to analyse
            splitting_conds (dict): Containing the following keys: \n
                'threshold':  
                    `(float)` The threshold at which data points of 
                    type 'value' within this CellBox are checked to be either 
                    above or below

        Returns:
            str:
                The homogeniety condtion returned is of the form: \n
                'MIN' = the cellbox contains less than a minimum number of 
                data points \n
                'HET' = Threshold values defined in config are exceeded \n
                'CLR' = None of the HET conditions were triggered \n
        '''
        # Get length of dataset in bounds  
        if type(self.data) == pd.core.frame.DataFrame:
            num_dp = len(self.trim_datapoints(bounds))
        elif type(self.data) == xr.core.dataset.Dataset:
            num_dp = min(self.trim_datapoints(bounds).count().values())

        # Set default homogeneity 
        hom_type = 'CLR'

        # Check to see if it's above the minimum threshold
        if num_dp < self.min_dp:
            logging.debug(f"\t{num_dp} datapoints found for attribute '{self.data_name}' within bounds '{bounds}'")
            hom_type = 'CLR'
        else:
            # To allow multiple modes of splitting, chuck them in the splitting conditions
            # Split if magnitude of curl(data) is larger than threshold 
            if 'curl' in splitting_conds:
                curl = self.calc_curl(bounds)
                if np.abs(curl) > splitting_conds['curl']:
                    hom_type =  'HET'
            # Split if max magnitude(any_vector - ave_vector) is larger than threshold
            if 'dmag' in splitting_conds:
                dmag = self.calc_dmag(bounds)
                if np.abs(dmag) > splitting_conds['dmag']:
                    hom_type = 'HET'
                
            # Split if Reynolds number is larger than threshold
            if 'reynolds' in splitting_conds:        
                reynolds = self.calc_reynolds_number(bounds)
                if reynolds > splitting_conds['reynolds']:
                    hom_type = 'HET'

        logging.debug(f"\thom_condition for attribute: '{self.data_name}' in bounds:'{bounds}' returned '{hom_type}'")
        
        return hom_type

    def reproject(self, in_proj='EPSG:4326', out_proj='EPSG:4326', 
                        x_col='lat', y_col='long'):
        '''
        Reprojects data using pyProj.Transformer
        self.data can be pd.DataFrame or xr.Dataset
        
        Args:
            in_proj (str): 
                Projection that the imported dataset is in
                Must be allowed by PyProj.CRS (Coordinate Reference System)
            out_proj (str): 
                Projection required for final data output
                Must be allowed by PyProj.CRS (Coordinate Reference System)
                Shouldn't change from default value (EPSG:4326)
            x_col (str): Name of coordinate column 1
            y_col (str): Name of coordinate column 2
                x_col and y_col will be cast into lat and long by the 
                reprojection 
            
        Returns:
            pd.DataFrame: 
                Reprojected data with 'lat', 'long' columns 
                replacing 'x_col' and 'y_col'
        '''
        def reproject_df(data, in_proj, out_proj, x_col, y_col):
            '''
            Reprojects a pandas dataframe
            
            Args:
                data (pd.DataFrame):
                    Data to reproject, with coordinates x_col, y_col
                in_proj (str): 
                    Projection the original dataset is in, as a string 
                    understandable by PyProj (e.g. 'EPSG:3031')
                out_proj (str):
                    Projection desired as output format. Should be always be
                    Mercator, but doesn't have to be if you desire
                    (e.g. 'EPSG:4326)
                x_col (str): 
                    Coordinate that original dataset includes that will be 
                    projected. Will be replaced with longitude values
                y_col (str):
                    Coordinate that original dataset includes that will be 
                    projected. Will be replaces with latitude values
                
            Returns:
                pd.DataFrame:
                    Reprojected dataset, with columns 'lat', 'long', 
                    ('time' if in original dataset), and data_name
            '''
            # Do the reprojection
            x, y = Transformer\
                    .from_crs(CRS(in_proj), CRS(out_proj), always_xy=True)\
                    .transform(data[x_col].to_numpy(), data[y_col].to_numpy())
            # Replace columns with reprojected columns called 'lat'/'long'
            if x_col != 'long':  data = data.drop(x_col, axis=1)
            if y_col != 'lat': data = data.drop(y_col, axis=1)
            data['lat']  = y
            data['long'] = x
            
            return data
            
        def reproject_xr(data, in_proj, out_proj, x_col, y_col, fast=False):
            '''
            Reprojects a xr.Dataset
            
            Args:
                data (xr.Dataset):
                    Data to reproject, with coordinates x_col, y_col
                in_proj (str): 
                    Projection the original dataset is in, as a string 
                    understandable by rioxarray (e.g. 'EPSG:3031')
                out_proj (str):
                    Projection desired as output format. Should be always be
                    Mercator, but doesn't have to be if you desire
                    (e.g. 'EPSG:4326)
                x_col (str): 
                    Coordinate that original dataset includes that will be 
                    projected. Will be replaced with longitude values
                y_col (str):
                    Coordinate that original dataset includes that will be 
                    projected. Will be replaces with latitude values
                
            Returns:
                pd.DataFrame:
                    Reprojected dataset, with columns 'lat', 'long', 
                    ('time' if in original dataset), and data_name
            '''
            if fast:
            # If want fast results (uses interpolation)
                max_size = sum(data.sizes.values())
                # Set data CRS
                data = data.rio.write_crs(in_proj)
                # Reproject
                data = data.rio.reproject(out_proj, resampling=Resampling.bilinear,
                                                    shape=((max_size,max_size)), 
                                                    nodata=np.nan)
                # Rename coordinates
                data = data.rename({x_col: 'long', y_col: 'lat'})
                # Reorder coords in case they are wrong
                data = data.sortby('lat', ascending=True)
                data = data.sortby('long', ascending=True)
                return data
            # If want accurate results
            else:
                df = data.to_dataframe().reset_index().dropna()
                return reproject_df(df, in_proj, out_proj, x_col, y_col)

        # If no reprojection to do
        if in_proj == out_proj:
            logging.debug("\tself.reproject() called but don't need to")
            return self.data
        else:
            logging.info(f"\tReprojecting data from {in_proj} to {out_proj}")
        # Choose appropriate method of reprojection based on data type
        if type(self.data) == pd.core.frame.DataFrame:
            return reproject_df(self.data, in_proj, out_proj, x_col, y_col)
        elif type(self.data) == xr.core.dataset.Dataset:
            return reproject_xr(self.data, in_proj, out_proj, x_col, y_col,
                                fast=self.fast_reprojection)
    
    def downsample(self, agg_type=None):
        '''
        Downsamples imported data to be more easily manipulated. Data size 
        should be reduced by a factor of m*n, where (m,n) are the 
        downsample_factors defined in the params.        
        self.data can be pd.DataFrame or xr.Dataset
        
        Args:
            agg_type (str): 
                Method of aggregation to bin data by to downsample. Default is
                same method used for homogeneity condition.            

        Returns:
            xr.Dataset or pd.DataFrame: 
                Downsampled data
        '''
        def downsample_xr(data, ds, agg_type):
            '''
            Downsample xarray dataset according to aggregation type
            
            Args:
                data (xr.Dataset):
                    Dataset containing data to be downsampled. Must have 
                    coordinates 'lat' and 'long'
                ds (int, int):
                    Downsampling factors. 
                    ds[0] is longitude
                    ds[1] is latitude
                agg_type (str):
                    Aggregation method to use for binning. Default is same as 
                    set in config, passed in by parent
            
            Returns:
                xr.Dataset:
                    Downsampled data
            '''
            if agg_type == 'MIN':
                # Returns min of bin
                data = data.coarsen(lat=ds[1],boundary='pad').min()
                data = data.coarsen(long=ds[0],boundary='pad').min()
            elif agg_type == 'MAX':
                # Returns max of bin
                data = data.coarsen(lat=ds[1],boundary='pad').max()
                data = data.coarsen(long=ds[0],boundary='pad').max()
            elif agg_type == 'MEAN':
                # Returns mean of bin
                data = data.coarsen(lat=ds[1],boundary='pad').mean()
                data = data.coarsen(long=ds[0],boundary='pad').mean()
            elif agg_type == 'MEDIAN':
                # Returns median of bin
                data = data.coarsen(lat=ds[1],boundary='pad').median()
                data = data.coarsen(long=ds[0],boundary='pad').median()
            elif agg_type == 'STD':
                # Returns std_dev of range
                data = data.coarsen(lat=ds[1],boundary='pad').std()
                data = data.coarsen(long=ds[0],boundary='pad').std()
            elif agg_type =='COUNT': 
                # Returns every first element in bin
                data = data.thin(lat=ds[1])
                data = data.thin(long=ds[0])
            return data
    
        def downsample_df(data, ds, agg_type):
            '''
            Downsample pandas dataframe
            Not implemented as it just adds to processing time, 
            defeating the purpose
            '''
            logging.warning(
                '\tDownsampling called on pd.DataFrame! Downsampling a df' \
                'too computationally expensive, returning original df'
                )
            return data

        # Set to params if no specific aggregate type specified
        if agg_type is None:
            agg_type = self.aggregate_type
        
        # If no downsampling
        if self.downsample_factors == (1,1) or \
           self.downsample_factors == [1,1]:
            logging.debug("\tself.downsample() called but don't have to")
            return self.data
        else:
            logging.info(f"\tDownsampling data by {self.downsample_factors}")
        # Otherwise, downsample appropriately
        if type(self.data) == pd.core.frame.DataFrame:
            return downsample_df(self.data, self.downsample_factors, agg_type)
        elif type(self.data) == xr.core.dataset.Dataset:
            return downsample_xr(self.data, self.downsample_factors, agg_type)
        
    def get_data_col_name(self):
        '''
        Retrieve name of data column (for pd.DataFrame), or variable 
        (for xr.Dataset). Used for when data_name not defined in params.
        Variable names are appended and comma seperated

        Returns:
            str: 
                Name of data columns, comma seperated
        '''
        def get_data_names_from_df(data):
            '''
            Filters out standard columns to extract only data column's name
            '''
            # Store name of data column for future reference
            columns = data.columns
            # Filter out lat, long, time columns leaves us with data column name
            filtered_cols = filter(lambda col: \
                                    col not in ['lat','long','time'], columns)
            data_names = list(filtered_cols)
            # Turn into comma seperated string and return
            return ','.join(data_names)
        
        def get_data_names_from_xr(data):
            '''
            Extracts variable name directly from xr.Dataset metadata
            '''
            # Extract data variables from xr.Dataset
            data_names = list(data.keys())
            # Turn into comma seperated string and return
            return ','.join(data_names)
        
        logging.debug(f"\tRetrieving data name from {type(self.data)}")
        # Choose method of extraction based on data type
        if type(self.data) == pd.core.frame.DataFrame:
            return get_data_names_from_df(self.data)
        elif type(self.data) == xr.core.dataset.Dataset:
            return get_data_names_from_xr(self.data)

    def get_data_col_name_list(self):
        '''
        Retrieve names of data columns (for pd.DataFrame), or variable 
        (for xr.Dataset). Used for when data_name not defined in params.

        Returns:
            list: 
                Contains strings of data namesk
        '''
        return self.get_data_col_name().split(',')

    def set_data_col_name(self, new_names):
        '''
        Sets name of data column/data variables from a comma-seperated string
        
        Args:
            name_dict (dict): 
                Dictionary mapping old variable names to new variable names,
                of the form {old_name (str): new_name (str)}

        Returns:
            xr.Dataset or pd.DataFrame: 
                Data with variable name changed
        '''
        def set_names_df(data, name_dict):
            '''
            Renames data columns in pandas dataframe
            '''
            # Rename data column to new name
            return data.rename(columns=name_dict)
        def set_names_xr(data, name_dict):
            '''
            Renames data variables in xarray dataset
            '''
            # Rename data variable to new name
            return data.rename(name_dict)
        # Get existing column names
        old_names = self.get_data_col_name().split(',')
        # Ensure that can do replacement of columns
        assert len(old_names) == len(new_names)
        # Set up mapping of old names to new names
        name_dict = {old_col: new_names[i] 
                     for i, old_col in enumerate(old_names)}
        # Change names
        # Change data name depending on data type
        if type(self.data) == pd.core.frame.DataFrame:
            return set_names_df(self.data, name_dict)
        elif type(self.data) == xr.core.dataset.Dataset:
            return set_names_xr(self.data, name_dict)
        
    def set_data_col_name_list(self, new_names):
        '''
        Sets name of data column/data variables from a list of strings.
        Also updates self.data_name_list with new names from list
        
        Args:
            new_names (list):
                List of strings containing new variable names
        
        Returns:
            pd.DataFrame or xr.Dataset:
                Original dataset with data variables renamed
        '''
        # Check validity of input
        assert type(new_names) == list, f"'new_names' must be a list! Instead it is a {type(new_names)}"
        assert len(new_names) == 2, f"'new_names' must have a length of 2! Instead it has length {len(new_names)}"
        str_items = [isinstance(name, str) for name in new_names]
        assert all(str_items), f"'new_names' must be list of 'str'. Currently {sum(str_items)} / 2 are strings!"
        new_data_name = ','.join(new_names)
        
        # Set names
        logging.info(f'\tSetting data names to {new_names}')
        self.data_name_list = new_names
        return self.set_data_col_name(new_data_name)

    def calc_reynolds_number(self, bounds):
        '''
        Calculates an approximate Reynolds number from the mean vector velocity
        and cellbox size.
        
        CURRENTLY ASSUMES DENSITY AND VISCOSITY OF SEAWATER AT 4°C! 
        WILL NEED MINOR REWORKING TO INCLUDE DIFFERENT FLUIDS
        
        Args:
            bounds (Boundary): 
                Cellbox boundary to calculate characteristic length from
                
        Returns:
            float:
                Reynolds number of cellbox
        '''
        # Extract the speed
        velocity = self.get_value(bounds, agg_type='MEAN')
        speed = np.linalg.norm(list(velocity.values())) # Calculates magnitude
        # Extract the characteristic length
        length = bounds.calc_size()
        # Calculate the reynolds number and return
        logging.warning("\tReynold number used for splitting, this function assumes properties of ocean water!")
        return 1028 * 0.00167 * speed * length

    def calc_divergence(self, bounds, data=None, collapse=True, agg_type='MAX'):
        '''
        Calculates the divergence of vectors in a cellbox
        
        Args:
            bounds (Boundary):
                Cellbox boundary in which all relevant vectors are contained
            data (pd.DataFrame or xr.Dataset):
                Dataset with 'lat' and 'long' columns/dimensions with vectors
            collapes (bool): 
                Flag determining whether to return an aggregated value, or a 
                vector field (values for each individual vector).
            agg_type (str):
                Method of aggregation if collapsing value. 
                Accepts 'MAX' or 'MEAN'
        
        Returns:
            float or pd.DataFrame:
                float value of aggregated div if collapse=True, or
                pd.DataFrame of div vector field if collapse=False 

        Raises:
            ValueError: If agg_type is not 'MAX' or 'MEAN'
        '''
        if data is None:    dps = self.trim_datapoints(bounds, data=data)
        else:               dps = data
        
        # Create a meshgrid of vectors from the data
        vector_field = self._create_vector_meshgrid(dps, self.data_name_list)

        # Get component values for each vector
        fx, fy = vector_field[:, :, 0], vector_field[:, :, 1]
        # If not enough datapoints to compute gradient
        if 1 in fx.shape or 1 in fy.shape:
            logging.debug('\tUnable to compute gradient across cell for divergence calculation')
            div = np.nan
        else:
            # Compute partial derivatives
            dfx_dy = np.gradient(fx, axis=1)
            dfy_dx = np.gradient(fy, axis=0)
            # Compute curl
            div = dfy_dx + dfx_dy
        
        # If div is nan
        if np.isnan(div).all():
            logging.debug('\tAll NaN cellbox encountered')
            return np.nan
        # If want to collapse to max mag value, return scalar
        elif collapse:   
            if agg_type == 'MAX':       return max(np.nanmax(div), np.nanmin(div), key=abs)
            elif agg_type == 'MEAN':    return np.nanmean(div)
            else: 
                raise ValueError(f"agg_type '{agg_type}' not understood! Requires 'MAX' or 'MEAN'")
        # Else return field
        else:
            return div


    def calc_curl(self, bounds, data=None, collapse=True, agg_type='MAX'):
        '''
        Calculates the curl of vectors in a cellbox
        
        Args:
            bounds (Boundary):
                Cellbox boundary in which all relevant vectors are contained
            data (pd.DataFrame or xr.Dataset):
                Dataset with 'lat' and 'long' columns/dimensions with vectors
            collapes (bool): 
                Flag determining whether to return an aggregated value, or a 
                vector field (values for each individual vector).
            agg_type (str):
                Method of aggregation if collapsing value. 
                Accepts 'MAX' or 'MEAN'
        
        Returns:
            float or pd.DataFrame:
                float value of aggregated curl if collapse=True, or
                pd.DataFrame of curl vector field if collapse=False
                
        Raises:
            ValueError: If agg_type is not 'MAX' or 'MEAN'
        '''
        if data is None:    dps = self.trim_datapoints(bounds, data=data)
        else:               dps = data
        # Create a meshgrid of vectors from the data
        vector_field = self._create_vector_meshgrid(dps, self.data_name_list)
        # Get component values for each vector
        fx, fy = vector_field[:, :, 0], vector_field[:, :, 1]
        # If not enough datapoints to compute gradient
        if 1 in fx.shape or 1 in fy.shape:
            logging.debug('\tUnable to compute gradient across cell for curl calculation')
            curl = np.nan
        else:
            # Compute partial derivatives
            dfx_dy = np.gradient(fx, axis=1)
            dfy_dx = np.gradient(fy, axis=0)
            # Compute curl
            curl = dfy_dx - dfx_dy

        # If div is nan
        if np.isnan(curl).all():
            logging.debug('\tAll NaN cellbox encountered')
            return np.nan
        # If want to collapse to max mag value, return scalar
        elif collapse:
            if agg_type == 'MAX': return max(np.nanmax(curl), np.nanmin(curl), key=abs)
            elif agg_type == 'MEAN':    return np.nanmean(curl)
            else: 
                raise ValueError(f"agg_type '{agg_type}' not understood! Requires 'MAX' or 'MEAN'")
        # Else return field
        else:
            return curl

    def calc_dmag(self, bounds, data=None, collapse=True, agg_type='MEAN'):
        '''
        Calculates the dmag of vectors in a cellbox.
        dmag is defined as being the difference in magnitudes between 
        each vector and the average vector within the bounds.\n
        dmag = mag(vector - mean_vector)
        
        Args:
            bounds (Boundary):
                Cellbox boundary in which all relevant vectors are contained
            data (pd.DataFrame or xr.Dataset):
                Dataset with 'lat' and 'long' columns/dimensions with vectors
            collapes (bool): 
                Flag determining whether to return an aggregated value, or a 
                vector field (values for each individual vector).
            agg_type (str):
                Method of aggregation if collapsing value. 
                Accepts 'MAX' or 'MEAN'
        
        Returns:
            float or pd.DataFrame:
                float value of aggregated dmag if collapse=True, or
                pd.DataFrame of dmag vector field if collapse=False
                
        Raises:
            ValueError: If agg_type is not 'MAX' or 'MEAN'
        '''
        if data is None:    dps = self.trim_datapoints(bounds, data=data)
        else:               dps = data
            
        data_names = self.data_name_list
        each_vector = dps[data_names].to_numpy()
        ave_vector = list(self.get_value(bounds, agg_type=agg_type).values())
        
        delta_vector = each_vector - ave_vector
        
        d_mag = np.linalg.norm(delta_vector, axis=1)
        if len(d_mag) == 0:
            logging.debug('\tEmpty cellbox encountered')
            return np.nan
        # If div is nan
        elif np.isnan(d_mag).all():
            logging.debug('\tAll NaN cellbox encountered')
            return np.nan
        # If want to collapse to max mag value, return scalar
        elif collapse:
            if agg_type == 'MAX':       return np.nanmax(d_mag)
            elif agg_type == 'MEAN':    return np.nanmean(d_mag)
            else:
                raise ValueError(f"agg_type '{agg_type}' not understood! Requires 'MAX' or 'MEAN'")
        # Else return field
        else:          return d_mag

    
    @staticmethod
    def _create_vector_meshgrid(data, data_name_list):
        '''
        Creates a np.meshgrid containing 2D vectors from a pd.DataFrame
        
        Args:
            data (pd.DataFrame): 
                Dataframe with columns 'lat', 'long', and vector x/y components
                lat | long | {v_x} | {v_y}
            data_name_list (list): 
                List of strings containing the vector component names
        
        Returns:
            np.array:
                Table containing vectors as np.arrays at each coord
        
        '''
        def meshgrid_from_df(data, data_name_list):

            # Manipulate into meshgrid of 2D vectors
            x, y = data_name_list
            # Fields of each vector component
            vector_x_field = data.pivot(index='lat', columns='long', values=x)
            vector_y_field = data.pivot(index='lat', columns='long', values=y)
            # Combine into field of vectors
            vector_field = np.stack((vector_x_field, vector_y_field), axis=-1)
            vector_field = np.swapaxes(vector_field, 0, 1)
            return vector_field
        
        def meshgrid_from_xr(data, data_name_list):
            # Extract out each variable and combine as tuple
            data_arrays = (data[name].values 
                           for name in data_name_list)
            # Zip them together to make 2D array of n-dimensional vectors
            return np.dstack(data_arrays)
        
        if type(data) == pd.core.frame.DataFrame:
            return meshgrid_from_df(data, data_name_list)
        elif type(data) == xr.core.dataset.Dataset:
            return meshgrid_from_xr(data, data_name_list)
        
        

    
    


from meshiphi.dataloaders.vector.abstract_vector import VectorDataLoader

import xarray as xr

class SOSEDataLoader(VectorDataLoader):
    def import_data(self, bounds):
        '''
        Reads in data from a SOSE Currents NetCDF file. 
        Renames coordinates to 'lat' and 'long', and renames variable to 
        'uC, vC'
        
        Args:
            bounds (Boundary): Initial boundary to limit the dataset to
            
        Returns:
            xr.Dataset: 
                SOSE currents dataset within limits of bounds. 
                Dataset has coordinates 'lat', 'long', and variable 'uC', 'vC'
        '''

        # Open dataset and cast to pandas df
        if len(self.files) == 1:    data = xr.open_dataset(self.files[0])
        else:                       data = xr.open_mfdataset(self.files)
        # Cast to dataframe to modify lon coordinate
        df = data.to_dataframe().reset_index()
        
        # Change long coordinate to be in [-180,180) domain rather than [0,360)
        df['long'] = df['lon'].apply(lambda x: x-360 if x>180 else x)
        # Extract relevant columns
        df = df[['lat','long','uC','vC']]
        # Trim to initial datapoints
        df = self.trim_datapoints(bounds, data=df)

        return df



from meshiphi.dataloaders.dataloader_interface import DataLoaderInterface
from abc import abstractmethod

from pyproj import Transformer, CRS

import logging
import numpy as np
import pandas as pd
from shapely.geometry import MultiPolygon, Point, LineString
from shapely.strtree import STRtree
from shapely.ops import unary_union

from meshiphi.utils import round_to_sigfig

class LutDataLoader(DataLoaderInterface):
    '''
    Abstract class for all LookUp Table Datasets.
    '''
    def __init__(self, bounds, params):
        '''
        This is where large-scale operations are performed, 
        such as importing data, downsampling, reprojecting, and renaming 
        variables
        
        Args:
            bounds (Boundary): 
                Initial mesh boundary to limit scope of data ingest
            params (dict): 
                Values needed by dataloader to initialise. Unique to each
                dataloader

        Attributes:
            self.data (gpd.DataFrame): 
                Data stored by dataloader to use when called upon by the mesh.
                Must be saved in mercator projection (EPSG:4326), with 
                columns 'geometry' and data_name.
            self.data_name (str): 
                Name of scalar variable. Must be the column name in the dataframe
                
        Raises:
            ValueError: If no data lies within the parsed boundary
        '''
        # Translates parameters from config input to desired inputs
        params = self.add_default_params(params)
        logging.info(f"Initialising {params['dataloader_name']} dataloader")
        # Creates a class attribute for all keys in params
        for key, val in params.items():
            setattr(self, key, val)
            
        # Read in and manipulate data to standard form
        self.data = self.import_data(bounds)
        if 'files' in params:
            logging.info('\tFiles read:')
            for file in self.files:
                logging.info(f'\t\t{file}')
        
        # Get data name from column name if not set in params
        if self.data_name is None:
            logging.debug('\tSetting self.data_name from column name')
            self.data_name = self.get_data_col_name()
        # or if set in params, set col name to data name
        else:
            logging.debug(f'\tSetting data column name to {self.data_name}')
            self.data = self.set_data_col_name(self.data_name)
            
        # Verify that all geometries are acceptable inputs
        self.data = self.verify_data()

        # Calculate fraction of boundary that data covers
        data_coverage = self.calculate_coverage(bounds)
        logging.info("\tMercator data range (roughly) covers "+\
                    f"{np.round(data_coverage*100,0).astype(int)}% "+\
                     "of initial boundary")
        # If there's 0 datapoints in the initial boundary, raise ValueError
        if data_coverage == 0:
            logging.error('\tDataloader has no data in initial region!')
            raise ValueError(f"Dataloader {params['dataloader_name']}"+\
                              " contains no data within initial region!")
        else:
            # Cut dataset down to initial boundary
            logging.info(
                "\tTrimming data to initial boundary: {min} to {max}".format(
                    min=(bounds.get_lat_min(), bounds.get_long_min()),
                    max=(bounds.get_lat_max(), bounds.get_long_max())
                ))
            
            self.data = self.trim_datapoints(bounds)
        
    @abstractmethod
    def import_data(self, bounds):
        '''
        User defined method for importing data from files, or even generating 
        data from scratch
                
        Returns:
            pd.DataFrame:
                Coordinates and data being imported from file \n
                if pd.DataFrame, 
                    - Must have columns 'geometry' and data_name
                    - Must have single data column
                    
        '''
        pass
        
    def add_default_params(self, params):
        '''
        Set default values for all LUT dataloaders. This function should be
        overloaded to include any extra params for a specific dataloader
        
        Args:
            params (dict): 
                Dictionary containing attributes that are required for each 
                dataloader. 
            
        Returns:
            (dict): 
                Dictionary of attributes the dataloader will require, 
                completed with default values if not provided in config.
        '''
        if 'dataloader_name' not in params:
            params['dataloader_name'] = self.__class__.__name__
        
        if 'data_name' not in params:
            params['data_name'] = None
            
        if 'aggregate_type' not in params: 
            params['aggregate_type']  = 'MEAN'
            
        return params
    
    def verify_data(self, data=None):
        """
        Verifies that all geometries read in are Polygons or MultiPolygons
        If MultiPolygon, then split out into multiple Polygons

        Args:
            data (pd.DataFrame, optional): 
                DataFrame with at least columns 'geometry' and a variable. 
                Defaults to dataloader's data attribute.

        Raises:
            ValueError: If read in a geometry that is not Polygon or MultiPolygon
        """
        
        if data is None:
            data = self.data
        
        # List of rows to include in cleaned dataframe
        new_data = []
        # For each entry
        for index, row in data.iterrows():
            # Add row if Polygon
            if row.geometry.geom_type == 'Polygon':
                new_data.append(data.loc[[index]])
            # Split out multipolygons to single polygons, and add rows
            elif row.geometry.geom_type == 'MultiPolygon':
                polygons = list(row.geometry.geoms)
                new_df = pd.concat([data.loc[[index]]]*len(polygons))
                new_df['geometry'] = polygons
                new_data.append(new_df)
            # Otherwise, geometry not usable
            else:
                raise ValueError(
                    f'{row.geometry.geom_type} not acceptable geometry type.'+\
                     ' Must be a Polygon or MultiPolygon!'
                )
        data = pd.concat(new_data, ignore_index=True)
        
        return data

    def calculate_coverage(self, bounds, data=None):
        """
        Calculates percentage of boundary covered by dataset

        Args:
            bounds (Boundary): 
                Boundary being compared against
            data (pd.DataFrame): 
                Dataset with shapely polygons in 'geometry' column 
                Defaults to objects internal dataset.
        
        Returns:
            float:
                Decimal fraction of boundary covered by the dataset
        """
        if data is None:
            data = self.data
        # Extract just polygons as a series if it's a dataframe
        if type(data) == pd.core.frame.DataFrame:
            data = data['geometry']
        # No coverage if no data
        if data.empty:
            return 0
        else:    
            # Calculate coverage fraction
            bounds_polygon = bounds.to_polygon()
            # Extract out polygons, add to multipolygon
            data_polygon = unary_union([datum for datum in data.tolist()])
            coverage = data_polygon.area / bounds_polygon.area

            # Cap output at 100%
            if coverage >= 1:   return 1
            else:               return coverage
        

    def trim_datapoints(self, bounds, data=None):
        '''
        Trims datapoints from self.data within boundary defined by 'bounds'.
        self.data can be pd.DataFrame or xr.Dataset
        
        Args:
            bounds (Boundary): Limits of lat/long/time to select data from

        Returns:
            pd.DataFrame: 
                Trimmed dataset in same format as self.data
        '''
        if data is None:
            data = self.data
        # Limit time to boundary
        if 'time' in data.index.names:
            data = data.loc[bounds.get_time_min():bounds.get_time_max()]
            
        # Find intersection of each polygon to the boundary
        bounds_polygon = bounds.to_polygon()    
        lut_polys = STRtree(list(data['geometry']))
        intersections = lut_polys.query(bounds_polygon, predicate='intersects').tolist()
        # Return only rows intersecting with cellbox boundary
        return data.iloc[intersections]

    def get_val_from_coord(self, long=None, lat=None, return_coords=False):
        '''
        Extracts value from self.data with lat and long specified in kwargs.
        
        Args:
            long (float): Longitude coordinate to search for
            lat (float) : Latitude coordinate to search for
            return_coords (boolean): 
                Flag for whether to return coordinates with the value or not
            
        Returns:
            pd.DataFrame: 
                Either with one entry (the value at the coords), 
                Optionaly with coordinates associated with it if 
                return_coords = True
        '''
        # Check if lat/long point is within any of the data polygons
        point = Point(long, lat)
        values = self.data[point.within(self.data['geometry'])][self.data_name]
        # Add lat/long coords to df if return_coords True
        if return_coords:
            data = pd.DataFrame({'lat': lat,
                                 'long': long,
                                 self.data_name: values})
        else:
            data = values
            
        return data
    
    def get_value(self, bounds, agg_type=None, skipna=False):
        '''
        Retrieve aggregated value from within bounds
        
        Args:
            aggregation_type (str): Method of aggregation of datapoints within
                bounds. Can be upper or lower case. 
                Accepts 'MIN', 'MAX', 'MEAN', 'MEDIAN', 'STD', 'COUNT'
            bounds (Boundary): Boundary object with limits of lat/long
            skipna (bool): Defines whether to propogate NaN's or not
                Default = False (includes NaN's)

        Returns:
            dict: 
                {variable (str): aggregated_value (float)}
                Aggregated value within bounds following aggregation_type
                
        Raises:
            ValueError: aggregation type not in list of available methods
        '''
        polygons = self.trim_datapoints(bounds)
        logging.debug(f"\t{len(polygons)} polygons found for attribute " + \
                      f"'{self.data_name}' within bounds '{bounds}'")
        
        ret_val = np.nan
        
        if agg_type is None:
            agg_type = self.aggregate_type
        # If want the number of datapoints
        if agg_type =='COUNT':
            ret_val = len(polygons)
        # If no data
        elif len(polygons) == 0:
            ret_val = np.nan
        # Return float of aggregated value
        elif agg_type == 'MIN':
            ret_val = polygons[self.data_name].min(skipna=skipna)
        elif agg_type == 'MAX':
            ret_val = polygons[self.data_name].max(skipna=skipna)
            
        elif agg_type in ['MEAN', 'MEDIAN', 'STD']:
            
            # Skipna not easily parsed to remaining calculations, so force it here
            if skipna: polygons = polygons.dropna()
            # Mean, median, and std dev need to be weighted by size of polygons
            polygons = polygons.assign(weights=lambda row: 
                                        self.calculate_coverage(bounds, 
                                                                row['geometry']))
            if agg_type == 'MEAN':
                ret_val = np.average(polygons[self.data_name], 
                                weights=polygons['weights'])
            
            elif agg_type == 'MEDIAN':
                # Chunk by cumulative weight
                polygons['cumulative_weights'] = polygons['weights'].cumsum(
                                                                    skipna=skipna)
                # Find middle of cumulative weight
                total_weight = polygons['weights'].sum(skipna=skipna)
                median_pos = total_weight/2
                # Extract top half and return lowest value (value closest to halfway)
                top_half = polygons.loc[polygons['cumulative_weights'] >= median_pos]
                ret_val = top_half.iloc[0]
            
            elif agg_type == 'STD':
                # Calculate std dev manually to account for weight
                average = np.average(polygons[self.data_name], 
                                    weights=polygons['weights'])
                variance = np.average(np.power(polygons[self.data_name]-average,2),
                                    weights=polygons['weights'])
                ret_val = np.sqrt(variance)
        # If aggregation_type not available
        else:
            raise ValueError(f'Unknown aggregation type {agg_type}')
        
        # Convert numpy bool to python bool for JSON serialisation
        if type(ret_val) is np.bool_:
            ret_val = bool(ret_val)
        
        return { self.data_name: ret_val}

    def get_hom_condition(self, bounds, splitting_conds):
        '''
        Retrieves homogeneity condition of data within
        boundary.
         
        Args: 
            bounds (Boundary): Boundary object with limits of datarange to analyse
            splitting_conds (dict): Containing the following keys: \n
                'boundary':  
                    `(boolean)` True if user wants to split when polygon boundary
                    goes through bounds 

        Returns:
            str:
                The homogeniety condtion returned is of the form: \n
                'CLR' = the boundary is completely contained within the LUT 
                regions, no need to split\n
                'MIN' = the boundary contains no LUT data, can't split \n
                'HET' = the boundary contains an edge within the LUT data, 
                should split
                
        '''
        bounds_polygon = bounds.to_polygon()
        # Extract polygons that overlap the boundary
        polygons = self.trim_datapoints(bounds)['geometry'].tolist()
        
    
        # If there's no polygon that overlaps with bounds        
        if not any([polygon.intersects(bounds_polygon) for polygon in polygons]):
            return 'CLR'
        # If we want to split on the boundary
        elif splitting_conds['boundary']:
            if any(p.boundary.intersects(bounds_polygon) for p in polygons):
                return 'HET'
            
        # Otherwise no boundaries intersected bounds
        return 'CLR'

    def reproject(self):
        '''
        Reprojection not supported by LookUpTable Dataloader
        '''
        logging.warning('Reprojection not supported by LookUpTable Dataloader')
    
    def downsample(self):
        '''
        Downsampling not supported by LookUpTable Dataloader
        '''
        logging.warning('Downsampling not supported by LookUpTable Dataloader')
        
    def get_data_col_name(self):
        '''
        Retrieve name of data column. 
        Used for when data_name not defined in params.

        Returns:
            str: 
                Name of data column
            
        Raises:
            AssertionError: 
                If multiple possible data columns found, can't retrieve data 
                name
        '''
        
        logging.debug(f"\tRetrieving data name from {type(self.data)}")
        
        unique_cols = list(set(self.data.columns) - set(['time','geometry']))
        
        assert(len(unique_cols) == 1), \
            f'Expected only one data column! Instead, found {unique_cols}'
        
        return unique_cols[0]

    def set_data_col_name(self, new_name):
        '''
        Sets name of data column/data variable
        
        Args:
            name (str): Name to replace currently stored name with

        Returns:
            pd.DataFrame: 
                Data with variable name changed
        '''
        
        old_name = self.get_data_col_name()
        if old_name != new_name:
            logging.info(f"\tChanging data name from {old_name} to {new_name}")
            return self.data.rename({old_name:new_name})
        else:
            logging.info(f"\tData is already labelled '{new_name}'")
            return self.data



"""

In this section we will discuss the usage of the MeshBuilder functionality of meshiphi.\n

Example:
    An example of how to run this code can be executed by running the following in an ipython/Jupyter Notebook::\n

        from meshiphi.mesh_generation.mesh_builder import MeshBuilder

        import json
        with open('./config.json', 'r') as f:
            config = json.load(f)['config']

        mesh_builder = MeshBuilder(config)
        mesh_builder.build_environmental_mesh() \n

"""

import logging
import math
import numpy as np

from tqdm import tqdm

from meshiphi.mesh_generation.jgrid_cellbox import JGridCellBox
from meshiphi.mesh_generation.boundary import Boundary
from meshiphi.mesh_generation.cellbox import CellBox
from meshiphi.mesh_generation.direction import Direction
from meshiphi.mesh_generation.environment_mesh import EnvironmentMesh
from meshiphi.mesh_generation.metadata import Metadata
from meshiphi.mesh_generation.neighbour_graph import NeighbourGraph
from meshiphi.mesh_generation.mesh import Mesh
from meshiphi.dataloaders.factory import DataLoaderFactory
from meshiphi.config_validation.config_validator import validate_mesh_config


class MeshBuilder:
    """

       A class resposible for building an environment mesh based on a provided config file.
       
    """

    def __init__(self, config):
        """

            Constructs a Mesh from a given config file.\n
            
            Args:
                config (dict): config file which defines the attributes of the Mesh to be constructed. config is of the form: \n
                        {\n
                        "region": {\n
                            "lat_min": (real),\n
                            "lat_max": (real),\n
                            "long_min": (real),\n
                            "long_max": (real),\n
                            "start_time": (string) 'YYYY-MM-DD',\n
                            "end_time": (string) 'YYYY-MM-DD',\n
                            "cell_width": (real),\n
                            "cell_height" (real),\n
                            "split_depth" (int)\n
                        },\n
                        "data_sources": [\n
                            {\n
                                "loader": (string)\n
                                "params" (dict)\n
                            }\n
                        ],\n
                        "splitting": { \n
                            "split_depth": (int),\n
                            "minimum_datapoints": (int)\n
                            }\n
                        }
                   
                    NOTE: In the case of constructing a global mesh, the longtitude range should be -180:180.

                    "j_grid" (bool): True if the Mesh to be constructed should be of the same format as the original Java CellGrid, to be used for regression testing.\n
                
        """
        logging.info("Initialising Mesh Builder")
        validate_mesh_config(config)
        self.config = config
        bounds = Boundary.from_json(config)

        # Configs may contain reference to system time for start_time and end_time
        # which are parsed to datetime format when initialising boundary.
        # updates config start_time/ end_time once system time has been parsed.
        self.config['region']['start_time'] = bounds.get_time_min()
        self.config['region']['end_time'] = bounds.get_time_max()

        
        cell_width = config['region']['cell_width']
        cell_height = config['region']['cell_height']

        self.validate_bounds(bounds, cell_width, cell_height)

        logging.info("Initialising mesh...")
        logging.info("Initialising cellboxes...")
     
        cellboxes = []
        cellboxes = self.initialize_cellboxes(bounds, cell_width, cell_height)

        grid_width = (bounds.get_long_max() -
                      bounds.get_long_min()) / cell_width

  

        min_datapoints = 5
        if 'splitting' in self.config:
            min_datapoints = self.config['splitting']['minimum_datapoints']
        meta_data_list = self.initialize_meta_data(bounds, min_datapoints)

        # checking to avoid any dummy cellboxes (the ones that was splitted and replaced)
        logging.info("Assigning data sources to cellboxes...")
        for cellbox in cellboxes:
            if isinstance(cellbox, CellBox):
                cellbox.set_minimum_datapoints(min_datapoints)
                # assign meta data to each cellbox
                cellbox.set_data_source(meta_data_list)

        
        logging.info("Initialising neighbour graph...")
        self.neighbour_graph = NeighbourGraph(cellboxes, grid_width)
        self.neighbour_graph.set_global_mesh (self.check_global_mesh(bounds, cellboxes, int(grid_width)))

        max_split_depth = 0
        if 'splitting' in self.config:
            max_split_depth = self.config['splitting']['split_depth']
        self.mesh = Mesh(bounds, cellboxes,
                         self.neighbour_graph, max_split_depth)
        self.mesh.set_config(config)
        if self.is_jgrid_mesh():
            logging.warning("We're using the legacy Java style cell grid")


    def initialize_meta_data(self, bounds, min_datapoints):
        meta_data_list = []
        splitting_conds = []
        if 'data_sources' in self.config.keys():
            for data_source in self.config['data_sources']:
                loader_name = data_source['loader']
                loader = DataLoaderFactory.get_dataloader(
                    loader_name, bounds, data_source['params'], min_datapoints)

                logging.debug("Creating data loader {}".format(
                    data_source['loader']))
                updated_splitting_cond = []  # create this list to get rid of the data_name in the conditions as it is not handeled by the DataLoader, remove after talking to Harry to address this in the loader
                if 'splitting_conditions' in data_source['params']:
                    splitting_conds = data_source['params']['splitting_conditions']
                    for split_cond in splitting_conds:
                        cond = split_cond[loader.data_name]
                        updated_splitting_cond.append(cond)

                value_fill_type = self.check_value_fill_type(data_source)

                # Update list of files in config to match the ones read in by dataloader
                if 'files' in data_source['params']:
                    data_source['params']['files'] = loader.files

                meta_data_obj = Metadata(
                    loader, updated_splitting_cond,  value_fill_type)
                meta_data_list.append(meta_data_obj)

        return meta_data_list
        
    def check_value_fill_type(self, data_source):
        def is_float(element: any) -> bool:
            if element is None: 
                    return False
            try:
                    float(element)
                    return True
            except ValueError:
                    return False
        value_fill_type = "parent"
        if 'value_fill_types' in data_source['params']:
            if is_float (data_source  ['params']['value_fill_types']) or   data_source['params']['value_fill_types'] in ["parent" ,"Nan"]:
                value_fill_type = data_source  ['params']['value_fill_types']
            else:
                logging.warning("Invalid value for value_fill_types, setting to the default(parent) instead.")
        return value_fill_type

    def is_jgrid_mesh(self):
        if 'j_grid' in self.config.keys():
            if  self.config['j_grid'] == "True":
                return True
        return False

    def initialize_cellboxes(self, bounds, cell_width, cell_height):
        cellboxes = []
        grid_width = (bounds.get_long_max() -
                      bounds.get_long_min()) / cell_width
        grid_height = (bounds.get_lat_max() -
                       bounds.get_lat_min()) / cell_height
        for lat in np.arange(bounds.get_lat_min(), bounds.get_lat_max(), cell_height):
            for long in np.arange(bounds.get_long_min(), bounds.get_long_max(), cell_width):
                cell_lat_range = [lat, lat+cell_height]
                cell_long_range = [long, long+cell_width]
                cell_bounds = Boundary(
                    cell_lat_range, cell_long_range, bounds.get_time_range())
                cell_id = str(len(cellboxes))
                if self.is_jgrid_mesh():
                    cellbox_indx = len(cellboxes)
                    cellbox = JGridCellBox(cell_bounds, cell_id)
                    x_coord = cellbox_indx % grid_width
                    y_coord = abs(math.floor(
                        cellbox_indx / grid_width) - (grid_height - 1))
                    cellbox.set_grid_coord(x_coord, y_coord)
                    cellbox.set_initial_bounds(cell_bounds)

                else:
                    cellbox = CellBox(cell_bounds, cell_id)
                cellboxes.append(cellbox)
        return cellboxes
    
    def add_dataloader(self, Dataloader, params, bounds=None, name='myDataLoader', min_dp = 5):
        '''
        Adds a dataloader to a pre-existing mesh by adding to the metadata
        
        Args:
            Dataloader (ScalarDataLoader or VectorDataLoader):
                Dataloader object to add to metadata
            params (dict):
                Parameters to initialise dataloader with
            bounds (Boundary):
            name (str):
                Name of the dataloader used in config
                
        Returns:
            MeshBuilder:
                Original MeshBuilder object (self) with added metadata for 
                new dataloader
        '''
        if bounds is None:
            bounds = Boundary.from_json(self.config)
        
        logging.debug('Adding dataloader')
        dataloader = Dataloader(bounds, params)
        updated_splitting_cond = []
        if 'splitting_conditions' in params:
            splitting_conds = params['splitting_conditions']
            updated_splitting_cond = [split_cond[dataloader.data_name] for split_cond in splitting_conds]

        data_source = {'loader': name,
                       'params': params}
        value_fill_type = self.check_value_fill_type(data_source)
        
        meta_data_obj = Metadata(
            dataloader, updated_splitting_cond,  value_fill_type)
        
        for cellbox in self.mesh.cellboxes:
            if isinstance(cellbox, CellBox):
                cellbox.set_minimum_datapoints(min_dp)
                # Add new meta data to list of data sources per cellbox
                cellbox.set_data_source(
                    cellbox.get_data_source() + [meta_data_obj]
                )

        

    def validate_bounds(self, bounds, cell_width, cell_height):
        assert (bounds.get_long_max() - bounds.get_long_min()) % cell_width == 0, \
            f"""The defined longitude region <{bounds.get_long_min()} :{bounds.get_long_max()}>
            is not divisable by the initial cell width <{cell_width}>"""

        assert (bounds.get_lat_max() - bounds.get_lat_min()) % cell_height == 0, \
            f"""The defined longitude region <{bounds.get_lat_min()} :{bounds.get_lat_max()}>
            is not divisable by the initial cell width <{cell_height}>"""

    def check_global_mesh(self, bounds , cellboxes, grid_width):
        """
            Checks if the mesh is a global one and connects the cellboxes at the minimum longtitude and max longtitude accordingly

           Args:
                bounds (Boundary): an object represents the bounds of the mesh
                cellboxes (list<Cellbox>): a list that contains the mesh initial cellboxes (before any splitting)
                grid_width (int): an int represents the width of the mesh ( the number of cellboxes it contains horizontally)
           Returns:
                is_global_mesh (bool): a boolean indicates if the mesh is a global one
        """
        is_global_mesh = False
        if bounds.get_long_max()== abs (bounds.get_long_min()) == 180: # check if it is a global mesh
            is_global_mesh = True
            # find indeces of cellboxes at the min longtitude and max longtitude 
            min_long_cellboxes = cellboxes [::grid_width]
            max_long_cellboxes = cellboxes [grid_width-1::grid_width]
            # update NG to connect cellboxes
            for i in range (0 , len(min_long_cellboxes)): 
                    self.neighbour_graph.add_neighbour (int (min_long_cellboxes[i].get_id()) , Direction.west, int (max_long_cellboxes[i].get_id()))
                    self.neighbour_graph.add_neighbour (int (max_long_cellboxes[i].get_id()) , Direction.east , int (min_long_cellboxes[i].get_id()))
                    # checks to avoid the very upper and lower cellboxes as they do not have north/south neighbours
                    if 0<= i < len(min_long_cellboxes)-1:
                        self.neighbour_graph.add_neighbour (int (min_long_cellboxes[i].get_id()) , Direction.north_west, int (max_long_cellboxes[i+1].get_id()))
                        self.neighbour_graph.add_neighbour (int (max_long_cellboxes[i].get_id()) , Direction.north_east, int (min_long_cellboxes[i+1].get_id()))
                    if 0<i<= len(min_long_cellboxes)-1: 
                        self.neighbour_graph.add_neighbour (int (min_long_cellboxes[i].get_id()) , Direction.south_west, int (max_long_cellboxes[i-1].get_id()))
                        self.neighbour_graph.add_neighbour (int (max_long_cellboxes[i].get_id()) , Direction.south_east, int (min_long_cellboxes[i-1].get_id()))
                   
        return is_global_mesh
    
    def to_json(self):
        """
            Returns this Mesh converted to a JSON object.

            Returns:
                json (json): a string representation of the CellGird parseable as a JSON object. The JSON object is of the form -

                    {
                        "config": the config used to initialize the Mesh,
                        "cellboxes": a list of CellBoxes contained within the Mesh,
                        "neighbour_graph": a graph representing the adjacency of CellBoxes within the Mesh
                    }
        """
        output = dict()
        output['config'] = {'mesh_info': self.config}
        output["cellboxes"] = self.mesh.get_cellboxes()
        output['neighbour_graph'] = self.neighbour_graph.get_graph()


    def split_and_replace(self, cellbox):
        """
            Replaces a cellbox given by parameter 'cellbox' in this grid with
            4 smaller cellboxes representing the four corners of the given cellbox.
            A neighbours map is then created for each of the 4 new cellboxes
            and the neighbours map for all surrounding cell boxes is updated.

            Args:
                cellbox (CellBox): the CellBox within this Mesh to be split into
                    4 smaller CellBox objects.

        """
        split_cellboxes = cellbox.split(len(self.mesh.cellboxes))
        self.mesh.cellboxes += split_cellboxes
        cellboxes = self.mesh.cellboxes
        cellbox_indx = cellboxes.index(cellbox)

        north_west_indx = cellboxes.index(split_cellboxes[0])
        north_east_indx = cellboxes.index(split_cellboxes[1])
        south_west_indx = cellboxes.index(split_cellboxes[2])
        south_east_indx = cellboxes.index(split_cellboxes[3])

        south_neighbour_indx = self.neighbour_graph.get_neighbours(
            cellbox_indx, 4)
        north_neighbour_indx = self.neighbour_graph.get_neighbours(
            cellbox_indx, -4)
        east_neighbour_indx = self.neighbour_graph.get_neighbours(
            cellbox_indx, 2)
        west_neighbour_indx = self.neighbour_graph.get_neighbours(
            cellbox_indx, -2)

        # Create neighbour map for SW split cell.
        sw_neighbour_map = {Direction.north_east: [north_east_indx],
                            Direction.east: [south_east_indx],
                            Direction.south_east: [],
                            Direction.south: [],
                            # update to use the NG class
                            Direction.south_west: self.neighbour_graph.get_neighbours(cellbox_indx, Direction.south_west),
                            Direction.west: [],
                            Direction.north_west: [],
                            Direction.north: [north_west_indx]}

        self.fill_sw_neighbour_map(
            south_west_indx, south_neighbour_indx, west_neighbour_indx, sw_neighbour_map)
        self.neighbour_graph.add_node(south_west_indx, sw_neighbour_map)

        # Create neighbour map for NW split cell
        nw_neighbour_map = {Direction.north_east: [],
                            Direction.east: [north_east_indx],
                            Direction.south_east: [south_east_indx],
                            Direction.south: [south_west_indx],
                            Direction.south_west: [],
                            Direction.west: [],
                            Direction.north_west: self.neighbour_graph.get_neighbours(cellbox_indx, Direction.north_west),
                            Direction.north: []}

        self.fill_nw_map(north_west_indx, north_neighbour_indx,
                         west_neighbour_indx, nw_neighbour_map)
        self.neighbour_graph.add_node(north_west_indx, nw_neighbour_map)

        # Create neighbour map for NE split cell
        ne_neighbour_map = {Direction.north_east: self.neighbour_graph.get_neighbours(cellbox_indx, Direction.north_east),
                            Direction.east: [],
                            Direction.south_east: [],
                            Direction.south: [south_east_indx],
                            Direction.south_west: [south_west_indx],
                            Direction.west: [north_west_indx],
                            Direction.north_west: [],
                            Direction.north: []}
        self.fill_ne_map(north_east_indx, north_neighbour_indx,
                         east_neighbour_indx, ne_neighbour_map)
        self.neighbour_graph.add_node(north_east_indx, ne_neighbour_map)

        # Create neighbour map for SE split cell
        se_neighbour_map = {Direction.north_east: [],
                            Direction.east: [],
                            Direction.south_east: self.neighbour_graph.get_neighbours(cellbox_indx, Direction.south_east),
                            Direction.south: [],
                            Direction.south_west: [],
                            Direction.west: [south_west_indx],
                            Direction.north_west: [north_west_indx],
                            Direction.north: [north_east_indx]}

        self.fill_se_map(south_east_indx, south_neighbour_indx,
                         east_neighbour_indx, se_neighbour_map)
        self.neighbour_graph.add_node(south_east_indx, se_neighbour_map)

        # Update neighbours of cellbox_indx with the new neighbours coming from splitting the cellbox.

        # Update north neighbours of cellbox_indx with the new neighbours comming from the splitted cellbox
        self.neighbour_graph.update_neighbours(
            cellbox_indx, [north_west_indx, north_east_indx], Direction.north, cellboxes)
        # Update east neighbour cellbox_indx with the new neighbours comming from the splitted cellbox
        self.neighbour_graph.update_neighbours(
            cellbox_indx, [north_east_indx, south_east_indx], Direction.east, cellboxes)
        # Update south neighbour cellbox_indx with the new neighbours comming from the splitted cellbox
        self.neighbour_graph.update_neighbours(
            cellbox_indx, [south_west_indx, south_east_indx], Direction.south, cellboxes)
        # Update west neighbour cellbox_indx with the new neighbours comming from the splitted cellbox
        self.neighbour_graph.update_neighbours(
            cellbox_indx, [north_west_indx, south_west_indx], Direction.west, cellboxes)
        # Update corner neighbour maps
        self.neighbour_graph.update_corner_neighbours(
            cellbox_indx, north_west_indx, north_east_indx, south_west_indx, south_east_indx)

        # remove the original splitted cellbox from the neighbour_graph
        self.neighbour_graph.remove_node(cellbox_indx)
        # set the original splitted cellbox to None
        cellboxes[cellbox_indx] = None

 ############################## methods to fill the neighbour maps of the splitted cells ########################
 
    def fill_se_map(self, south_east_indx, south_neighbour_indx, east_neighbour_indx, se_neighbour_map):
        """
             method that fills the South east neighbours
        """
        cellboxes = self.mesh.cellboxes
        for indx in south_neighbour_indx:
            if self.neighbour_graph.get_neighbour_case(cellboxes[south_east_indx], cellboxes[indx]) == Direction.south:
                se_neighbour_map[4].append(indx)
            if self.neighbour_graph.get_neighbour_case(cellboxes[south_east_indx], cellboxes[indx]) == Direction.south_west:
                se_neighbour_map[-1].append(indx)
        for indx in east_neighbour_indx:
            if self.neighbour_graph.get_neighbour_case(cellboxes[south_east_indx], cellboxes[indx]) == Direction.east:
                se_neighbour_map[2].append(indx)
            if self.neighbour_graph.get_neighbour_case(cellboxes[south_east_indx], cellboxes[indx]) == Direction.north_east:
                se_neighbour_map[1].append(indx)

   
    def fill_ne_map(self, north_east_indx, north_neighbour_indx, east_neighbour_indx, ne_neighbour_map):
        """
             method that fills the North east neighbours
        """
        cellboxes = self.mesh.cellboxes
        for indx in north_neighbour_indx:
            if self.neighbour_graph.get_neighbour_case(cellboxes[north_east_indx], cellboxes[indx]) == Direction.north:
                ne_neighbour_map[-4].append(indx)
            if self.neighbour_graph.get_neighbour_case(cellboxes[north_east_indx], cellboxes[indx]) == Direction.north_west:
                ne_neighbour_map[-3].append(indx)
        for indx in east_neighbour_indx:
            if self.neighbour_graph.get_neighbour_case(cellboxes[north_east_indx], cellboxes[indx]) == Direction.east:
                ne_neighbour_map[2].append(indx)
            if self.neighbour_graph.get_neighbour_case(cellboxes[north_east_indx], cellboxes[indx]) == Direction.south_east:
                ne_neighbour_map[3].append(indx)

    
    def fill_nw_map(self,  north_west_indx, north_neighbour_indx, west_neighbour_indx, nw_neighbour_map):
        """
             method that fills the North west neighbours
        """
        cellboxes = self.mesh.cellboxes
        for indx in north_neighbour_indx:
            if self.neighbour_graph.get_neighbour_case(cellboxes[north_west_indx], cellboxes[indx]) == Direction.north:
                nw_neighbour_map[-4].append(indx)
            if self.neighbour_graph.get_neighbour_case(cellboxes[north_west_indx], cellboxes[indx]) == Direction.north_east:
                nw_neighbour_map[1].append(indx)
        for indx in west_neighbour_indx:
            if self.neighbour_graph.get_neighbour_case(cellboxes[north_west_indx], cellboxes[indx]) == Direction.west:
                nw_neighbour_map[-2].append(indx)
            if self.neighbour_graph.get_neighbour_case(cellboxes[north_west_indx], cellboxes[indx]) == Direction.south_west:
                nw_neighbour_map[-1].append(indx)

  
    def fill_sw_neighbour_map(self, south_west_indx, south_neighbour_indx, west_neighbour_indx, sw_neighbour_map):
        """
             method that fills the South west neighbours
        """
        cellboxes = self.mesh.cellboxes
        for indx in south_neighbour_indx:
            if self.neighbour_graph.get_neighbour_case(cellboxes[south_west_indx], cellboxes[indx]) == Direction.south_east:
                sw_neighbour_map[3].append(indx)
            if self.neighbour_graph.get_neighbour_case(cellboxes[south_west_indx], cellboxes[indx]) == Direction.south:
                sw_neighbour_map[4].append(indx)
        for indx in west_neighbour_indx:
            if self.neighbour_graph.get_neighbour_case(cellboxes[south_west_indx], cellboxes[indx]) == Direction.west:
                sw_neighbour_map[-2].append(indx)
            if self.neighbour_graph.get_neighbour_case(cellboxes[south_west_indx], cellboxes[indx]) == Direction.north_west:
                sw_neighbour_map[-3].append(indx)


    def split_to_depth(self, split_depth):
        """
            splits all cellboxes in this grid until a maximum split depth
            is reached, or all cellboxes are homogeneous.

            Args:
                split_depth (int): The maximum split depth reached by any CellBox
                    within this Mesh after splitting.
        """
        logging.info ("Splitting cellboxes...")
        # loop over the data_sources then cellboxes to implement depth-first splitting. should be simpler and loop over cellboxes only once we switch to breadth-first splitting
        # this impl assumws all the cellboxes have the same data sources. should not be the caase once we switch to breadth-first splitting.
        data_sources = self.mesh.cellboxes[0].get_data_source()
        
        # Set up data_source progress bar
        ds_pbar = tqdm(range(0, len(data_sources)), position=0, 
                       bar_format='{desc}{n_fmt}/{total_fmt} |{bar}| {percentage:3.0f}%, [{elapsed} elapsed] ')
        sd_pbar = tqdm(range(0, split_depth), position=1, leave=False, 
                        bar_format=' Split depth: {n_fmt}/{total_fmt} |{bar}| {percentage:3.0f}%{postfix} ')
        for index in ds_pbar:
            # Update name of data source being processed
            ds_pbar.set_description(f' Processing {data_sources[index].get_data_loader().dataloader_name} data')
            
            if (len(data_sources[index].get_splitting_conditions()) > 0):
                # Set up split depth progress bar
                level = 0
                sd_pbar = tqdm(range(0, split_depth), position=1, leave=False, 
                               bar_format=' Split depth: {n_fmt}/{total_fmt} |{bar}| {percentage:3.0f}%{postfix} ')

                for cb_index, cellbox in enumerate(self.mesh.cellboxes):
                    ds_pbar.update(0)
                    if isinstance(cellbox, CellBox):
                        if cellbox.get_split_depth() > level:
                            level = cellbox.get_split_depth()
                            # If we're a split level further down, iterate progress in progress bar
                            sd_pbar.update()
                        # Split the cellbox
                        should_split = cellbox.should_split(index+1)
                        if (cellbox.get_split_depth() < split_depth) & should_split:
                            self.split_and_replace(cellbox)
                        # Update number of cellboxes processed
                        sd_pbar.set_postfix_str(f'[Cellbox {cb_index+1} / {len(self.mesh.cellboxes)}]')
        tqdm.write('')
        ds_pbar.clear()
        sd_pbar.clear()

    def build_environmental_mesh(self):
        """
            splits the mesh then goes through the mesh cellboxes and builds an evironmental mesh that contains the cellboxes aggregated data
            
            Returns:
                EnvironmentMesh: an object that represents the constructed nonunifrom mesh and contains the aggregated cellboxs and neighbour graph 
        """
        self.split_to_depth(self.mesh.get_max_split_depth())
        agg_cellboxes = []

        agg_cell_count = 0
        logging.info('Aggregating cellboxes...')
        for cellbox in tqdm(self.mesh.cellboxes, 
                            bar_format=' Aggregating cellboxes: {n_fmt}/{total_fmt} |{bar}| {percentage:3.0f}%, [{elapsed} elapsed] '):
            agg_cell_count += 1
            if isinstance(cellbox, CellBox):
                logging.debug(f'aggregating cellbox ({agg_cell_count}/{len(self.mesh.cellboxes)})')
                agg_cellboxes.append(cellbox.aggregate())

        env_mesh = EnvironmentMesh(self.mesh.get_bounds(
        ), agg_cellboxes, self.neighbour_graph, self.get_config())

        return env_mesh

    def get_config(self):
        """
        returns the config
        """
        return self.config








import numpy as np
from meshiphi.mesh_generation.boundary import Boundary
from meshiphi.mesh_generation.aggregated_cellbox import AggregatedCellBox
import logging
class JGridAggregatedCellBox (AggregatedCellBox):
    """
    a class represnts an aggrgated information within a geo-spatial/temporal boundary. 


    Attributes:
      

    Note:
        All geospatial boundaries of a CellBox are given in a 'EPSG:4326' projection
    """
    

    def __init__(self, boundary , agg_data , id):
        """

            Args:
                boundary(Boundary): encapsulates latitude, longtitude and time range of the CellBox
        """
        # Box information relative to bottom left
        self.boundary = boundary
        self.agg_data = agg_data
        self.id = id
        self.node_string =""
        
######## setters and getters ########
    def set_node_string (self , str):
       self.node_string = str

    def get_node_string (self ):
       return self.node_string

    def set_Boundary(self, boundary):
        """
            set the boundary of the CellBox
        """
        self.boundary = boundary

    def set_agg_data (self, agg_data):
      
        self.agg_data = agg_data
    
    def set_id (self, id):
      
        self.id = id
    
    def get_bounds(self):
        """
            get the boundary of the CellBox
        """
        return self.boundary 

    def get_agg_data (self):
      
        return self.agg_data

    def get_id (self):
      
        return self.id

    def mesh_dump(self):
        logging.info ("dumping the mesh ...")
        """
            returns a string representing all the information stored in the mesh
            of this cellbox

            for use in j_grid regression testing
        """
        number_of_points=0
        mesh_dump = ""
        mesh_dump += self.get_node_string() + "; "  # add node string
        mesh_dump += "0 "
        ice_area = 0
        uc = None
        vc = None
        mesh_dump += str(self.boundary.getcy()) + ", " + str(self.boundary.getcx()) + "; "  # add lat,long
     
        value = self.agg_data ['SIC']      
        number_of_points = self.agg_data["SIC_COUNT"]
        if value != None:
            ice_area = value 
      
        uc = self.agg_data['uC']*3.6  #unit conversion to match the Java code
        vc = self.agg_data['vC']*3.6 #unit conversion to match the Java code 

        mesh_dump += str(ice_area) + "; "  # add ice area
        # add uc, uv
        if np.isnan(uc) or np.isnan(vc): 
            uc = 0.0
            vc = 0.0 
        
        mesh_dump += str(uc) + ", " + str(vc) + ", "
        mesh_dump += str(number_of_points) 
        mesh_dump += "\n"

        return mesh_dump


"""
Outlined in this section we will discuss the usage of the CellBox functionality of the meshiphi package.
In this series of class distributions we house our discrete representation of input data. In each CellBox,
we represent a way of accessing the information governing our numerical world,
this includes and is not limited to: Ocean Currents, Sea Ice Concentration and Bathymetric depth.\n


    Example:\n
    An example of running this code can be executed by running the following in a ipython/Jupyter Notebook:: \n

            from meshiphi.mesh_generation.cellbox import cellbox \n
            .... \n

    Note:\n
        CellBoxes are intended to be constructed by and used within a Mesh object.
          The methods provided are to extract information for CellBoxes contained within a Mesh. \n

"""


import numpy as np
from meshiphi.mesh_generation.boundary import Boundary
from meshiphi.mesh_generation.aggregated_cellbox import AggregatedCellBox
import logging



class CellBox:
    """
    A CellBox represnts a geo-spatial/temporal boundary that enables projecting to information within.
    Information about any given value of a CellBox is calculated from aggregating all data points of within those bounds.
    CellBoxes may  be split into smaller CellBoxes and the data points within distributed  between the newly created
    CellBoxes so as to construct a non-uniform mesh of CellBoxes, such as within a Mesh.\n

    Attributes:
        Bounds (Boundary): object that contains the latitude and logtitute range and the time range \n
        id (int):  the id of the cellbox \n

    """

    def __init__(self, bounds, id):
        """

            Args:
                bounds(Boundary): encapsulates latitude, longtitude and time range of the CellBox\n
                id (int):  the id of the cellbox \n
        """
        # Box information relative to bottom left
        self.bounds = bounds
        self.parent = None
        self.minimum_datapoints = 10
        self.split_depth = 0
        self.data_source = None
        self.id = id

######## setters and getters ########
    def set_minimum_datapoints(self, minimum_datapoints):
        """
            set the minimum number of data contained within CellBox boundaries
        """
        if minimum_datapoints < 0:
            raise ValueError(
                f'CellBox: minimum number of data contained can not be negative')
        self.minimum_datapoints = minimum_datapoints

    def set_data_source(self, data_source):
        """
            a method that sets the data source of the cellbox ( which includes the data loaders,
              splitting conditions and aggregation type)

            Args:
                data_source (List <MetaData>): a list of MetaData objects, each object represents a source of this CellBox data
                  (where the data comes from, how it is spitted and aggregated)
        """
        self.data_source = data_source

    def set_parent(self, parent):
        """
            set the parent CellBox, which is the bigger CellBox that conains this CellBox
            Args:
                CellBox: the bigger Cellbox object that got splitted to produce this cellbox
        """
        self.parent = parent

    def set_split_depth(self, split_depth):
        """
            set the split depth of a CellBox, which represents is the number of times the CellBox
            has been split to reach it's current size.
        """
        if split_depth < 0:
            raise ValueError(f'CellBox: split depth can not be negative')
        self.split_depth = split_depth

    def set_id(self, id):
        """
        method ssts cellbox id
        """
        self.id = id

    def get_id(self):
        """
        method returns cellbox cell id
        """
        return self.id

    def get_minimum_datapoints(self):
        """
            get the minimum number of data contained within CellBox boundaries
        """
        return self.minimum_datapoints

    def get_data_source(self):
        """
            a method that gets the data source of the cellbox
              (the data loaders, splitting conditions and aggregation type)
            returns:
            data_source (List <MetaData>): a list of MetaData objects, each object represents a source
            of this CellBox data (where the data comes from, how it is spitted and aggregated)
        """
        return self.data_source

    def get_parent(self):
        """
            get the parent CellBox, which is the bigger CellBox that conains this CellBox
        """
        return self.parent

    def get_bounds(self):
        """
            get the spatial and temporal bounds (lat range, long range and time range) of this cellbox
        """
        return self.bounds

    def get_split_depth(self):
        """
            get the split depth of a CellBox, which represents is the number of times the CellBox
              has been split to reach it's current size.
        """
        return self.split_depth

######################################################
# methods used for splitting a cellbox

    def should_split(self, stop_index):
        """
            determines if a cellbox should be split based on the homogeneity
            condition of each data type contained within. The homogeneity condition
            of values within this cellbox is calculated using the method
            'get_hom_cond' in each DataLoader object inside CellBox's metadata\n

            if ANY data returns 'HOM':
                do not split
            if ANY data returns 'MIN':
                do not split
            if ALL data returns 'CLR':
                do not split
            else (mixture of CLR & HET):
                split\n
            Args:
                stop_index: the index of the data source at which checking the splitting conditions stops.
                Implemented like this to perform depth-first splitting.
                Should be deprecated once we switch to breadth-first splitting
            Returns:
                 bool: True if the splitting_conditions of this CellBox
                    will result in the CellBox being split.
        """
        hom_conditions = []

        current_data_source = None
        for index in range(0, stop_index):
            current_data_source = self.get_data_source()[index]
            data_loader = current_data_source.get_data_loader()
            for splitting_cond in current_data_source.get_splitting_conditions():
                hom_cond = data_loader.get_hom_condition(
                    self.bounds, splitting_cond)
                hom_conditions.append(hom_cond)

        if "HOM" in hom_conditions:
            return False
        if "MIN" in hom_conditions:
            return False
        if hom_conditions.count("CLR") == len(hom_conditions):
            return False

        return True

    def should_split_breadth_first(self):
        """
            determines if a cellbox should be split based on the homogeneity
            condition of each data type contained within. The homogeneity condition
            of values within this cellbox is calculated using the method
            'get_hom_cond' in each DataLoader object inside CellBox's metadata

            if ANY data returns 'HOM':
                do not split
            if ANY data returns 'MIN':
                do not split
            if ALL data returns 'CLR':
                do not split
            else (mixture of CLR & HET):
                split

            Returns:
                should_split (bool): True if the splitting_conditions of this CellBox
                    will result in the CellBox being split.
        """
        hom_conditions = []
        for current_data_source in self.data_source:
            data_loader = current_data_source.get_data_loader()
            for splitting_cond in current_data_source.get_splitting_conditions():
                hom_cond = data_loader.get_hom_condition(
                    self.bounds, splitting_cond)
                hom_conditions.append(hom_cond)

        if "HOM" in hom_conditions:
            return False
        if "MIN" in hom_conditions:
            return False
        if hom_conditions.count("CLR") == len(hom_conditions):
            return False

        return True

    def split(self, start_id):
        """
            splits the current cellbox into 4 corners, returns as a list of cellbox objects.

            Args:
                start_id : represenst the start of the splitted cellboxes ids,
                  usuallly it is the number of the existing cellboxes
            Returns:
                 list<CellBox>: The 4 corner cellboxes generated by splitting
                 the cellbox uniformly.
        """
        split_boxes = self.create_splitted_cell_boxes(start_id)

        # set CellBox split_depth, data_source and parent
        for split_box in split_boxes:
            split_box.set_split_depth(self.get_split_depth() + 1)
            split_box.set_data_source(self.get_data_source())
            split_box.set_parent(self)

        return split_boxes

    def create_splitted_cell_boxes(self, index):
        """
        method that creates 4 splitted cellbox
        """
        half_width = self.bounds.get_width() / 2
        half_height = self.bounds.get_height() / 2

        # create 4 new cellboxes
        time_range = self.bounds.get_time_range()
        lat = self.bounds.get_lat_min()
        lat_range = [lat + half_height, lat + self.bounds.get_height()]
        long = self.bounds.get_long_min()
        long_range = [long, long + half_width]
        boundary = Boundary(lat_range, long_range, time_range)
        north_west = CellBox(boundary, str(index))

        lat_range = [lat + half_height, lat + self.bounds.get_height()]
        long_range = [long + half_width, long + self.bounds.get_width()]
        boundary = Boundary(lat_range, long_range, time_range)
        index += 1
        north_east = CellBox(boundary, str(index))

        lat_range = [lat, lat + half_height]
        long_range = [long, long + half_width]
        boundary = Boundary(lat_range, long_range, time_range)
        index += 1
        south_west = CellBox(boundary, str(index))

        lat_range = [lat, lat + half_height]
        long_range = [long + half_width, long + self.bounds.get_width()]
        boundary = Boundary(lat_range, long_range, time_range)
        index += 1
        south_east = CellBox(boundary, str(index))

        split_boxes = [north_west, north_east, south_west, south_east]
        return split_boxes

    def aggregate(self):
        '''
            aggregates CellBox data using the associated data_sources' aggregate type (ex. MEAN, MAX)
            and returns AggregatedCellBox object

            Returns:
                AggregatedCellbox: object contains the aggregated data within cellbox bounds.
        '''
        agg_dict = {}
        for source in self.get_data_source():
            loader = source.get_data_loader()
            # get the aggregated value from the associated DataLoader
            agg_value = loader.get_value(self.bounds)
            data_name = loader.data_name
            parent = self.get_parent()
            # check if the data name has many entries (ex. uC,uV)
            if ',' in data_name:
                agg_value = self.check_vector_data(
                    source, loader, agg_value, data_name)

            elif np.isnan(agg_value[data_name]):
                if source.get_value_fill_type() == 'parent':
                    # if the agg_value empty and get_value_fill_type is parent, then use the parent bounds
                    while parent is not None and np.isnan(agg_value[data_name]):
                        agg_value = loader.get_value(parent.bounds)
                        parent = parent.get_parent()
                else:  # not parent, so either float or Nan so set the agg_Data to value_fill_type
                    agg_value[data_name] = source.get_value_fill_type()
                    
            # If already and value for {data_name} in cellbox and it's a NaN, overwrite
            if (data_name in agg_dict):
                if np.isnan(agg_dict[data_name]):
                    logging.warning(f'\t{data_name} already exists in cellbox! Overwriting only NaN values')
                    agg_dict.update(agg_value)
            else:
                agg_dict.update(agg_value)
                
        agg_cellbox = AggregatedCellBox(self.bounds, agg_dict, self.get_id())
        # free the memory space used by the cellbox
        self.deallocate_cellbox()
        return agg_cellbox

    def check_vector_data(self, source, loader, agg_value, data_name):
        """
        method that checks if the vector data is None and calls the parent get value
        """
        data_name_list = data_name.split(',')
        for name in data_name_list:
            parent = self.get_parent()
            if np.isnan(agg_value[name]):
                if source.get_value_fill_type() == 'parent':
                    # if the agg_value empty and get_value_fill_type is parent, then use the parent bounds
                    while parent is not None and np.isnan(agg_value[name]):
                        agg_value[name] = loader.get_value(parent.bounds)[name]
                        parent = parent.get_parent()
                else:  # not parent, so either float or Nan so set the agg_Data to value_fill_type
                    agg_value[data_name] = source.get_value_fill_type()
        return agg_value

    def deallocate_cellbox(self):
        """
        Method to free up the memory space allocated by the cellbox
        """
        for source in self.get_data_source():
            loader = source.get_data_loader()
            del loader
            del source
        parent = self.parent
        # free up the memory space used by the parent cellboxes chain
        while isinstance(parent, CellBox):
            for source in self.parent.get_data_source():
                loader = source.get_data_loader()
                del loader
                del source
            grandparent = parent.parent
            del parent
            parent = grandparent
        del self



from meshiphi.dataloaders.vector.abstract_vector import VectorDataLoader

import logging

import xarray as xr
import numpy as np

#TODO Read in 2 files, combine to one object
class ORAS5CurrentDataLoader(VectorDataLoader):
    def import_data(self, bounds):
        '''
        Reads in data from a ORAS5 Depth NetCDF files. 
        Renames coordinates to 'lat' and 'long', and renames variable to 
        'uC, vC'
        
        Args:
            bounds (Boundary): Initial boundary to limit the dataset to
            
        Returns:
            xr.Dataset: 
                ORAS5 currents dataset within limits of bounds. 
                Dataset has coordinates 'lat', 'long', and variable 'uC', 'vC'
        '''
        # Open Dataset
        if len(self.files) == 1:    data = xr.open_dataset(self.files[0])
        else:                       data = xr.open_mfdataset(self.files)
        
        # Change column names
        data = data.rename({'nav_lon': 'long',
                            'nav_lat': 'lat',
                            'uo': 'uC',
                            'vo': 'vC'})
        # Limit to just these coords and variables
        data = data[['lat','long','uC','vC']]
        
        # Limit to initial boundary
        data = self.trim_datapoints(bounds, data=data)
        
        return data


from meshiphi.dataloaders.scalar.abstract_scalar import ScalarDataLoader

import logging
import xarray as xr

class GEBCODataLoader(ScalarDataLoader):
    def import_data(self, bounds):
        '''
        Reads in data from GEBCO NetCDF files. Renames coordinates to
        'lat' and 'long'.
        
        Args:
            bounds (Boundary): Initial boundary to limit the dataset to
            
        Returns:
            xr.Dataset: 
                GEBCO dataset within limits of bounds. 
                Dataset has coordinates 'lat', 'long', and variable 'elevation'         
        '''
        # Import data from files defined in config
        if len(self.files) == 1:    data = xr.open_dataset(self.files[0])
        else:                       data = xr.open_mfdataset(self.files)
        # Rename columns to standard format
        data = data.rename({'lon':'long'})
        # Trim to initial datapoints
        data = self.trim_datapoints(bounds, data=data)
        
        return data






from datetime import datetime
from datetime import timedelta

from math import cos, sin, asin, sqrt, radians
from shapely import wkt

class Boundary:
    """
    A Boundary is a class that defines the geo-spatial/temporal
    boundaries (longtitude, latitude and time).


    Attributes:
        lat_range (float[]): array contains the start and end of latitude range 
        long_range (float[]): array contains the start and end of longtitude range.
          In the case of constructing a global mesh, the longtitude range should be -180:180.
        time_range(string[]): array contains the start and end of time range 
        

    Note:
        All geospatial boundaries are given in a 'EPSG:4326' projection
    """ 
    @classmethod
    def from_json(cls, config):
        """
             constructs a boundary object from json input
            Args:
               config (json): json object that contains the boundary attributes
                
        """
        long_min = config['region']['long_min']
        long_max = config['region']['long_max']
        lat_min = config['region']['lat_min']
        lat_max = config['region']['lat_max']
        start_time = config['region']['start_time']
        end_time = config['region']['end_time']
        lat_range = [lat_min, lat_max]
        long_range = [long_min , long_max]
        time_range = [start_time , end_time]
        obj = Boundary (lat_range , long_range , time_range)
        return obj



    def __init__(self, lat_range , long_range , time_range=None):
        """

            Args:
               lat_range (float[]): array contains the start and end of latitude range 
               long_range (float[]): array contains the start and end of longtitude range 
               time_range(Date[]): array contains the start and end of time range 
                
        """
        if time_range is None:
            time_range=[]
        else: 
             time_range[0] = self.parse_datetime(time_range[0])
             time_range[1] = self.parse_datetime(time_range[1])

        self.validate_bounds(lat_range , long_range , time_range)
        # Boundary information 
        self.lat_range = lat_range
        self.long_range = long_range
        self.time_range = time_range

    def parse_datetime(self, datetime_str: str):
        """
            Attempts to parse a string containing reference to system time into datetime format.
            If given the string 'TODAY', will return system time.
            special characters '+' and '-' can be used to adjust system time. e.g 'TODAY + 3' 
            will return system time + 3 days, 'TODAY - 16' will return system time - 16 days
            
            Args:
                datetime_str (String): String attempted to be parsed to datetime format. 
                    Expected input format is '%d%m%Y'
            Returns:
                date (String): date in a String format, '%Y-%m-%d'.
            Raises:
                ValueError : If given 'datetime_str' cannot be parsed, raises ValueError.
        """
        DATE_FORMAT = "%Y-%m-%d"
        
        datetime_str = datetime_str.upper()

        # check if datetime_str is valid datetime format.
        try:
            datetime_object = datetime.strptime(datetime_str, DATE_FORMAT).date()
            return datetime_object.strftime(DATE_FORMAT)
        except ValueError:
            # check if datetime_str contains reference to system-time.
            if datetime_str.strip() == "TODAY":
                today = datetime.today()
                return today.strftime(DATE_FORMAT)
            elif "TODAY" not in datetime_str:
                raise ValueError(f'Incorrect date format given. Cannot convert "{datetime_str}" to date.')
            
                # check for increment to system time.
            if "+" in datetime_str:
                increment = datetime_str.split("+")[1].strip()
                try:
                    increment = float(increment)
                    datetime_object = datetime.today() + timedelta(days = increment)
                    return datetime_object.strftime(DATE_FORMAT)
                except ValueError:
                    raise ValueError(f'Incorrect date format given. Cannot convert "{datetime_str}" to date. ' + \
                        f'Time increment "{increment}" cannot be cast to float')
                
                # check for decrement to system time.
            elif "-" in datetime_str:
                decrement = datetime_str.split("-")[1].strip()
                try:
                    decrement = float(decrement)
                    datetime_object = datetime.today() - timedelta(days = decrement)
                    return datetime_object.strftime(DATE_FORMAT)
                except ValueError:
                    raise ValueError(f'Incorrect date format given. Cannot convert "{datetime_str}" to date. ' + \
                            f'Time decrement "{decrement}" cannot be cast to float')
            else:
                raise ValueError(f'Incorrect date format given. Cannot convert "{datetime_str}" to date')
        

    def validate_bounds (self, lat_range , long_range , time_range):
        """
            method to check the bounds are valid
            Args:
               lat_range (float[]): array contains the start and end of latitude range 
               long_range (float[]): array contains the start and end of longtitude range 
               time_range(Date[]): array contains the start and end of time range 
                
        """
        if len(lat_range) < 2 or len (long_range)<2 :
            raise ValueError('Boundary: range should contain two values')
        if lat_range[0] > lat_range [1]:
             raise ValueError('Boundary: Latitude start range should be smaller than range end')
        if long_range[0] > long_range [1]:
             raise ValueError('Boundary: Longtitude start range should be smaller than range end')
        if long_range[0] < -180 or long_range[1] > 180:
            raise ValueError('Boundary: Longtitude range should be within -180:180')
        if len (time_range) > 0:
            if datetime.strptime(time_range[0], '%Y-%m-%d') > datetime.strptime(time_range[1], '%Y-%m-%d'):
                     raise ValueError('Boundary: Start time range should be smaller than range end')

    # Functions used for getting data from a cellbox
    def getcx(self):
        """
            returns x-position of the centroid of the cellbox

            Returns:
                cx (float): the x-position of the top-left corner of the CellBox
                    given in degrees longitude.
        """
        return self.long_range[0] + self.get_width()/2
    def getcy(self):
        """
            returns y-position of the centroid of the cellbox

            Returns:
                cy (float): the y-position of the top-left corner of the CellBox
                    given in degrees latitude.
        """ 
        return self.lat_range[0] + self.get_height()/2
    def get_height(self):
        """
            returns height of the cellbox

            Returns:
                height (float): the height of the CellBox
                    given in degrees latitude.
        """
        height = self.lat_range[1] - self.lat_range[0]
        return height
    def get_width(self):
        """
            returns width of the cellbox

            Returns:
                width (float): the width of the CellBox
                    given in degrees longtitude.
        """
        width = self.long_range[1] - self.long_range[0]
        return width
    def get_time_range (self):
        """
            returns the time range
        """
        return self.time_range
    def getdcx(self):
        """
            returns x-distance from the edge to the centroid of the cellbox

            Returns:
                dcx (float): the x-distance from the edge of the CellBox to the 
                    centroid of the CellBox. Given in degrees longitude
        """
        return self.get_width()/2
    def getdcy(self):
        """
            returns y-distance from the edge to the centroid of the cellbox

            Returns:
                dxy (float): the y-distance from the edge of the CellBox to the
                    centroid of the CellBox. Given in degrees latitude
        """
        return self.get_height()/2
    def get_lat_min(self):
        """
            returns the min latitude
        """
        return self.lat_range[0]
    def get_lat_max(self):
        """
            returns the max latitude
        """
        return self.lat_range[1]   
    def get_long_min(self):
        """
            returns the min longtitude
        """
        return self.long_range[0]
    def get_long_max(self):
        """
            returns the max longtitude
        """
        return self.long_range[1]  
    def get_time_min(self):
        """
            returns the min of time range
        """
        return self.time_range[0]

    def get_time_max(self):
        """
            returns the max of time range
        """

        return self.time_range[1] 
    def get_bounds(self):
        """
            returns the bounds of this cellbox

            Returns:
                bounds (list<tuples>): The geo-spatial boundaries of this CellBox.
        """
        bounds = [[ self.long_range[0], self.lat_range[0] ],
                   [ self.long_range[0], self.lat_range[1]],
                    [ self.long_range[1], self.lat_range[1]],
                    [ self.long_range[1], self.lat_range[0]],
                    [self.long_range[0], self.lat_range[0], ]]
        return bounds
    
    def calc_size(self):
        """
        Calculate the great circle distance (in meters) between 
        two points on the earth (specified in decimal degrees)
        """
        # convert decimal degrees to radians
        lon1, lat1, lon2, lat2 = map(radians, [self.get_long_min(), 
                                               self.get_lat_min(),
                                               self.get_long_max(),
                                               self.get_lat_max()])
        # haversine formula
        dlon = lon2 - lon1
        dlat = lat2 - lat1
        a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
        c = 2 * asin(sqrt(a))
        # Get diagonal length
        m = (6371 * c * 1000)
        # Divide by sqrt(2) to get 'square' side length
        return m / sqrt(2)
    
    def to_polygon(self):
        """
        Creates a shapely polygon from the extent of the boundary. Will be a 
        rectangle in mercator projection.
        
        Returns:
            (shapely.geometry.polygon.Polygon):
                Shapely polygon with corners at the min/max lat/long 
                values of this boundary
        """
        polygon = wkt.loads(
                    f'POLYGON(({self.get_long_min()} {self.get_lat_min()},' + \
                             f'{self.get_long_min()} {self.get_lat_max()},' + \
                             f'{self.get_long_max()} {self.get_lat_max()},' + \
                             f'{self.get_long_max()} {self.get_lat_min()},' + \
                             f'{self.get_long_min()} {self.get_lat_min()}))'
            )
        return polygon

    def __str__(self):


        lat_range = "lat range :[" + str(self.get_lat_min()) + \
              "," + str(self.get_lat_max()) + "]"
        long_range = "long range :[" + str(self.get_long_min()) + \
              "," + str(self.get_long_max()) + "]"
        time_range = "time range :" + str(self.get_time_range())

        return "{"+ lat_range + ", " + long_range + ", " + time_range + "}"





region_schema = {
    "type": "object",
    "required": ["lat_min",  "lat_max", 
                 "long_min", "long_max",
                 "start_time", "end_time",
                 "cell_width", "cell_height"],
    "additionalProperties": False,
    "properties":{    
        "lat_min": {"type": "number"},
        "lat_max": {"type": "number"},
        "long_min": {"type": "number"},
        "long_max": {"type": "number"},
        "start_time": {"type": "string"},
        "end_time": {"type": "string"},
        "cell_width": {"type": "number"},
        "cell_height": {"type": "number"}
    }
}

dataloader_schema = {
    "type": "object",
    "required": ["loader", "params"],
    "additionalProperties": False,
    "properties":{
        "loader": {"type": "string"},
        "params": {"type": "object"}
        
    }
}

splitting_schema = {
    "type": "object",
    "required": ["split_depth", "minimum_datapoints"],
    "additionalProperties": False,
    "properties":{
        "split_depth": {"type": "integer"},
        "minimum_datapoints": {"type": "integer"}
    } 
}

mesh_schema = {
    "type": "object",
    "required": ["region", "data_sources", "splitting"],
    "properties":{
        "region": {
            "$ref": "#/region_schema"
        },
        "data_sources": {
            "type": "array",
            "items": {
                "$ref": "#/dataloader_schema"
            },
        },
        "splitting": {
            "$ref": "#/splitting_schema"
        }
    },
    "region_schema": region_schema,
    "dataloader_schema": dataloader_schema,
    "splitting_schema": splitting_schema
}




from meshiphi.mesh_validation.sampler import Sampler


from meshiphi.mesh_generation.mesh_builder import MeshBuilder
from meshiphi.mesh_generation.mesh import Mesh
import numpy as np
import json
import math
import logging
from meshiphi.mesh_generation.boundary import Boundary
from sklearn.metrics import mean_squared_error
import xarray as xr
class MeshValidator:

    """
    a class that validates a constructed mesh against its actual sourced geo-spatial data. Validation takes place by comparing the aggregated data value of mesh's cellbox against the actual data contained within cellbox's bounds.
    Attributes:
        conf (dict): conatins the initial config used to build the mesh under vlaidation
        validation_length (float): the legnth of the validation square used for each sample, all the data_points contained within this square wil be validated. Higher values would incur higher processing cost
        mesh (Mesh): object that represents the constructed mesh (a representation before aggregating cellbox data, used to have access to the mesh data source to validate against the actual data)
        env_mesh (EnvironmentMesh): objects that represents the constructed env mesh (a representation after aggregating the mesh cellboox data)

    """

    def __init__ (self , mesh_config_file):
        """

            Args:
              mesh_config_file (String): the path to the config file used to build the mesh under validation

        """
        self.conf = None
        self.validation_length = .1 
        with open (mesh_config_file , "r") as config_file:
            self.conf = json.load(config_file)['config']['mesh_info']

        mesh_builder = MeshBuilder (self.conf)
        self.env_mesh =  mesh_builder.build_environmental_mesh()
        self.mesh = mesh_builder.mesh

    
       

    def validate_mesh (self , number_of_samples=10):
        """

          samples the mesh's lat and long space and compares the actual data within the sampled's range to the mesh agg_value then calculates the RMSE.

            Args:
              number_of_samples (int): the number of samples used to validate the mesh
            Returns:
                distance (float): the RMSE between the actaul data value and the mesh's agg_value.

        """
        
        # read the mesh bounds then generate samples of lat and long within bounds
        SAMPLE_DIM = 2  # each sample contains lat and long
    
        bounds = self.mesh.get_bounds()
        samples = Sampler(SAMPLE_DIM , number_of_samples).generate_samples([bounds.lat_range , bounds.long_range])
        # compare the sampled lat and long values in data_file to the values obtained by mesh ( agg_values returned by  get_value)
        actual_value = np.array([])
        mesh_value = np.array([])
        for sample in samples:
           actual_value =  np.append (actual_value ,self.get_value_from_data (sample))
           mesh_value =  np.append ( mesh_value ,self.get_values_from_mesh(sample))  
        # calculate the RMSE over the samples.
        distance = math.sqrt (mean_squared_error(actual_value,mesh_value))
        return distance
    

    def get_value_from_data (self , sample):
        """
            gets the actual data within the provided sample lat and long
            Args:
              sample (float[]): a decimal array contains the sampled lat and long values
            Returns:
                a numpy array that contains all the data within the sampled lat and long range
        """
        values =[]
        #calculate the sampling range based on the validation length
        lat_end, long_end = self.get_range_end(sample)
        lat_range = [sample[0] , lat_end]
        long_range = [sample[1] , long_end ]
        time_range = self.mesh.get_bounds().get_time_range()
        for source in self.mesh.cellboxes[0].get_data_source():
            data_loader = source.get_data_loader() 
            data_name = data_loader.get_data_col_name()
            dp = data_loader.trim_datapoints( Boundary (lat_range , long_range , time_range))
            values = np.append (values , dp[data_name])
        logging.info("values from data are: {}".format(' '.join(map(str, values))))
        return values

    def get_range_end(self, sample):
        """
            calculates the range end of the provided sample lat and long, claculation is based on the specified validation_length
            Args:
              sample (float[]): a decimal array contains the sampled lat and long values
            Returns:
                float[]: lat and long range end
        """
        lat_end = sample[0] + self.validation_length
        long_end = sample[1] + self.validation_length
        # make sure we are not exceeding the mesh bounds
        if lat_end > self.mesh.get_bounds().get_lat_max():
            lat_end = self.mesh.get_bounds().get_lat_max()
        if long_end > self.mesh.get_bounds().get_long_max():
            long_end = self.mesh.get_bounds().get_long_max()
        return lat_end,long_end


    def get_values_from_mesh (self , sample):
            
            """
                finds the mesh's cellboxes that contains the sample's lat and long then returns the aggregated values within.
                Args:
                sample (float[]): a decimal array contains the sampled lat and long values
                Returns:
                    a numpy array that contains the mesh's data within the sampled lat and long range
            """
            #TODO make sure to handle the vector data
            values = []
            #calculate the sampling range based on the validation length
            lat_end, long_end = self.get_range_end(sample)
            lat_range = [sample[0] , lat_end]
            long_range = [sample[1] , long_end ]
            time_range = self.mesh.get_bounds().get_time_range()

        
            for source in self.mesh.cellboxes[0].get_data_source():
                data_loader = source.get_data_loader()
                dp = data_loader.trim_datapoints( Boundary (lat_range , long_range , time_range))
                
                if type(dp) == xr.core.dataset.Dataset:
                    dp = dp.to_dataframe().reset_index()
                for index, point in dp.iterrows():
                    lat = point ['lat']
                    long = point ['long']
                    for agg_cellbox in self.env_mesh.agg_cellboxes:
                        if agg_cellbox.contains_point(lat , long):
                            values = np.append ( values , agg_cellbox.agg_data [data_loader.data_name] )#get the agg_value 
                            break  # break to make sure we avoid getting multiple values (for lat and long on bounds of multiple cellboxes)
            logging.info("values from mesh are: {}".format(' '.join(map(str, values))))
         
            return values
    

        
        






from meshiphi.dataloaders.scalar.abstract_scalar import ScalarDataLoader
from meshiphi.utils import gaussian_random_field

import logging
import pandas as pd
import numpy as np

class ScalarGRFDataLoader(ScalarDataLoader):
    
    def add_default_params(self, params):
        '''
        Set default values for abstract GRF dataloaders, starting by
        including defaults for scalar dataloaders.
        
        Args:
            params (dict): 
                Dictionary containing attributes that are required for the
                shape being loaded. Must include 'shape'.
            
        Returns:
            (dict): 
                Dictionary of attributes the dataloader will require, 
                completed with default values if not provided in config.
        '''
        params = super().add_default_params(params)
        
        # Params that all GRF dataloaders need
        if 'seed' not in params:
            params['seed'] = None
        if 'size' not in params:
            params['size'] = 512
        if 'alpha' not in params:
            params['alpha'] = 3
        
        # Set defaults for binary GRF
        if params['binary'] == True:
            if params['data_name'] is None:
                params['data_name'] = 'binary_data'
            if 'min' not in params:
                params['min'] = 0
            if 'max' not in params:
                params['max'] = 1
            # If threshold not set, make it average of min/max val
            if 'threshold' not in params:
                params['threshold'] = 0.5
        # Set defaults for smooth scalar GRF
        else:
            if params['data_name'] is None:
                params['data_name'] = 'scalar_data'
            if 'min' not in params:
                params['min'] = -10
            if 'max' not in params:
                params['max'] = 10
            # If threshold not set, make it min/max vals
            if 'threshold' not in params:
                params['threshold'] = [0,1]
            if 'multiplier' not in params:
                params['multiplier'] = 1
            if 'offset' not in params:
                params['offset'] = 0
                
        return params
        
    
    def import_data(self, bounds):
        '''
        Creates data in the form of a Gaussian Random Field
        
        Args:
            bounds (Boundary): Initial boundary to limit the dataset to
            
        Returns:
            pd.DataFrame: 
                Scalar dataset within limits of bounds. Dataset has coordinates
                'lat', 'long', and variable name 'data'. \n
                NOTE - In __init__, variable name will be set to data_name if 
                defined in config
                
        '''
    
        def grf_to_binary(grf, threshold):
            '''
            Creates a mask out of a GRF if params specify a binary output
            '''
            # Cast all above threshold to 1, below to 0
            grf[grf >= threshold] = 1.0
            grf[grf <  threshold] = 0.0
            # Cast 0/1 to False/True
            grf = (grf == True)
            return grf
        
        def grf_to_scalar(grf, threshold, min_val, max_val):
            # Bound GRF to threshold limits
            grf[grf <  threshold[0]] = threshold[0]
            grf[grf >= threshold[1]] = threshold[1]
            # Renormalise the GRF
            grf = grf - np.min(grf)
            grf = grf/(np.max(grf)-np.min(grf))
            # Scale to max/min
            grf = grf * (max_val - min_val) + min_val
            
            return grf
        
        # Set seed for generation. If not specified, will be 'random'
        np.random.seed(self.seed)
        
        # Create a GRF
        grf = gaussian_random_field(self.size, self.alpha)
        
        # Set it to a binary mask if chosen in config
        if self.binary == True:
            grf = grf_to_binary(grf, self.threshold)
        # Set it to scalar GRF field if chosen in config
        else:
            grf = grf_to_scalar(grf, self.threshold, self.min, self.max)
            # Scale with multiplier and offset
            grf = self.multiplier * grf + self.offset
    
        # Set up domain of field
        lat_array = np.linspace(bounds.get_lat_min(), 
                                bounds.get_lat_max(), 
                                self.size)
        long_array = np.linspace(bounds.get_long_min(), 
                                 bounds.get_long_max(), 
                                 self.size)
        latv, longv = np.meshgrid(lat_array, long_array, indexing='ij')

        # Create an entry for each row in final dataframe
        rows = [
            {'lat': latv[i,j], 'long': longv[i,j], 'data': grf[i,j]}
            for i in range(self.size) 
            for j in range(self.size)
            ]
        # Cast to dataframe
        data = pd.DataFrame(rows).set_index(['lat','long'])
        # Set to xarray dataset
        data = data.to_xarray()
        
        return data


vessel_schema = {
    "type": "object",
    "required": ["VesselType", "MaxSpeed",
                 "Unit", "MaxIceConc",
                 "MinDepth"],
    "additionalProperties": True,
    "properties": {
        "VesselType": {"type": "string"},
        "MaxSpeed": {"type": "number", "minimum": 0},
        "Unit": {"type": "string"},
        "MaxIceConc": {"type": "number", "minimum": 0, "maximum": 100},
        "MinDepth": {"type": "number", "minimum": 0}
    }
}



from meshiphi.dataloaders.lut.abstract_lut import LutDataLoader
from meshiphi.mesh_generation.boundary import Boundary
import logging

import geopandas as gpd
import pandas as pd
from shapely.ops import unary_union
import numpy as np

class LutGeoJSON(LutDataLoader):
        
    def add_default_params(self, params):
        '''
        Set default values for LUT dataloader. Only unique addition over
        the regular abstracted add_default_params is the data_name
        
        Args:
            params (dict): 
                Dictionary containing attributes that are required for the
                LUT being loaded.
            
        Returns:
            (dict): 
                Dictionary of attributes the dataloader will require, 
                completed with default values if not provided in config.
        '''
        # Add default scalar params
        params = super().add_default_params(params)
        
        # Name of data being read in from GeoJSONs
        if 'data_name' not in params:
            params['data_name'] = 'dummy_data'
            
        return params
    
    
    def import_data(self, bounds):
        """
        Import a list of GeoJSON files, assign regions a value specified in
        config params, regions outside this are numpy nan values.
        
        Args:
            bounds (Boundary): Initial boundary to limit the dataset to
            
        Returns:
            exclusion_df (pd.DataFrame): 
                Dataframe of polygons with value specified in config.
                DataFrame has columns 'geometry' and 
                data_name ('dummy_data' by default)
        """
        # Read in all files specified and extract geometry of geojsons
        gdf_list = [gpd.read_file(file) for file in self.files]
        geojson_df = gpd.GeoDataFrame(pd.concat(gdf_list, ignore_index=True))
        # Give geometries a value defined in config
        geojson_df[self.data_name] = self.value
        geojson_df = pd.DataFrame(geojson_df[['geometry',self.data_name]])
        
        # Create boundary denoting the world 
        world_polygon = Boundary([-90, 90], [-180, 180]).to_polygon()
        # Subtract out all regions with defined values
        defined_polygon = unary_union(geojson_df.geometry)
        undefined_polygon = world_polygon - defined_polygon
        # Set remainder to have value np.nan (or opposing bool value)
        if self.value is True:      alt_val = False
        elif self.value is False:   alt_val = True
        else:                       alt_val = np.nan
        geojson_df.loc[len(geojson_df.index)] = [undefined_polygon, alt_val]
        # Limit to boundary
        geojson_df = self.trim_datapoints(bounds, data=geojson_df)

        return geojson_df


from meshiphi.dataloaders.scalar.abstract_scalar import ScalarDataLoader
from datetime import datetime
import logging

import xarray as xr
class AMSRDataLoader(ScalarDataLoader):
    
    def add_default_params(self, params):
        '''
        Translates 'hemisphere' parameter into values of in_proj and out_proj 
        that pyProj can understand. Also defines x_col and y_col for AMSR data
        for reprojection to function
        
        Args:
            params (dict): 
                Dictionary holding keys and values that will be turned into 
                object attributes
        
        Returns:
            dict:
                Params dictionary with addition of translated key/value pairs
        '''
        # Set default parameters same as all other scalar dataloaders
        params = super().add_default_params(params)
        # Translate 'hemisphere' into initial projection
        hemisphere = params['hemisphere'].lower()
        if  hemisphere == 'north':
            params['in_proj'] = 'EPSG:3411'
        elif hemisphere == 'south':
            params['in_proj'] = 'EPSG:3412'
        else:
            raise ValueError(
                "Hemisphere defined in config is not 'north' or 'south'"
                )
        # Set initial projection column names
        params['x_col'] = 'x'
        params['y_col'] = 'y'
        
        return params
            
    def import_data(self, bounds):
        '''
        Reads in data from a AMSR NetCDF file, or folder of files. 
        Drops unnecessary column 'polar_stereographic', and renames variable
        'z' to 'SIC'
        
        Args:
            bounds (Boundary): Initial boundary to limit the dataset to
            
        Returns:
            xr.Dataset: 
                AMSR SIC dataset within limits of bounds. 
                Dataset has coordinates 'lat', 'long', and variable 'SIC'
        '''
        def retrieve_date(filename):
            '''
            Get date from filename in format:
                asi-AMSR2-s6250-<year><month><day>-v.5.4.nc
            '''
            date = filename.split('-')[-2]
            date = f'{date[:4]}-{date[4:6]}-{date[6:]}'
            return date
        
        def retrieve_data(filename, date):
            '''
            Read in data as xr.Dataset, create time coordinate
            '''
            data = xr.open_dataset(filename)
            # Add date to data
            data = data.assign_coords(time=date)
            return data

        data_array = []
        relevant_files = []
        # For each file found from config
        for file in self.files:
            # If date within boundary
            date = retrieve_date(file)
            # If file data from within time boundary, append to list
            # Doing now to avoid ingesting too much data initially
            if datetime.strptime(bounds.get_time_min(), '%Y-%m-%d') <= \
                datetime.strptime(date, '%Y-%m-%d') <= \
                datetime.strptime(bounds.get_time_max(), '%Y-%m-%d'):
                data_array.append(retrieve_data(file, date))
                relevant_files += [file]
        # Concat all valid files
        if len(data_array) == 0:
            logging.error('\tNo files found for date range '+\
                         f'[ {bounds.get_time_min()} : {bounds.get_time_max()} ]')
            raise FileNotFoundError('No AMSR files found within specified time range!')
        data = xr.concat(data_array,'time')

        # Remove unnecessary column, rename data column
        data = data.drop_vars('polar_stereographic')
        data = data.rename({'z': 'SIC'})
        
        # Limit self.files to only those actually used
        self.files = relevant_files
        
        # TODO Limit data range before reprojection
        
        return data



from meshiphi.dataloaders.vector.abstract_vector import VectorDataLoader

import logging
 
import xarray as xr

class BalticCurrentDataLoader(VectorDataLoader):
    def import_data(self, bounds):
        '''
        Reads in current data from a copernicus baltic sea physics reanalysis NetCDF file.
        Renames coordinates to 'lat' and 'long', and renames variable to 
        'uC, vC'
        
        Args:
            bounds (Boundary): Initial boundary to limit the dataset to
            
        Returns:
            xr.Dataset: 
                Baltic currents dataset within limits of bounds. 
                Dataset has coordinates 'lat', 'long', and variable 'uC', 'vC'
        '''
        # Open Dataset
        if len(self.files) == 1:    data = xr.open_dataset(self.files[0])
        else:                       data = xr.open_mfdataset(self.files)
        # Change column names
        data = data.rename({'latitude': 'lat',
                            'longitude': 'long',
                            'uo': 'uC',
                            'vo': 'vC'})
        
        # Trim to initial datapoints
        data = self.trim_datapoints(bounds, data=data)
        
        return data



from shapely.geometry import Polygon
from meshiphi.mesh_generation.boundary import Boundary
import shapely.wkt


class AggregatedCellBox:
    """
    a class represnts an aggrgated information within a geo-spatial/temporal boundary. 


    Attributes:
      

    Note:
        All geospatial boundaries of a CellBox are given in a 'EPSG:4326' projection
    """
    @classmethod
    def from_json(cls, cellbox_json):
        """

            Args:
                cellbox_json(Json): json object that encapsulates boundary, agg_data and id of the CellBox
        """
        cellbox_id = cellbox_json ['id']
        def load_boundary (cellbox_json):
        
            shape = shapely.wkt.loads (cellbox_json ["geometry"])
            bounds = shape.bounds
            lat_range = [bounds[1] , bounds[3]]
            long_range = [bounds [0], bounds [2]]
            return Boundary (lat_range , long_range)

        def load_agg_data (cellbox_json):
            dict_obj = {}
            for key in cellbox_json:
                if key  not in [  "geometry","cx", "cy", "dcx", "dcy"]:
                    dict_obj[key] = cellbox_json[key]

            return dict_obj
        
        boundary = load_boundary( cellbox_json)
        agg_data = load_agg_data( cellbox_json)
        obj = AggregatedCellBox(boundary , agg_data ,cellbox_id )
        return obj




    def __init__(self, boundary , agg_data , id):
        """

            Args:
                boundary(Boundary): encapsulates latitude, longtitude and time range of the CellBox
                agg_data (dict): a dictionary that contains data_names and agg values
                id (string): a string represents cellbox id
        """
        # Box information relative to bottom left
        self.boundary = boundary
        self.agg_data = agg_data
        self.id = id
        
######## setters and getters ########
    def set_boundary(self, boundary):
        """
            set the boundary of the CellBox
        """
        self.boundary = boundary

    def set_agg_data (self, agg_data):
        """
        sets the agg_data
        """
        self.agg_data = agg_data
    
    def set_id (self, id):
        """
        sets the cellbox id
        """
        self.id = id
    
    def get_bounds(self):
        """
            get the boundary of the CellBox
        """
        return self.boundary 

    def get_agg_data (self):
        """
        returns the agg_data
        """
        return self.agg_data

    def get_id (self):
        """
        returns the id
        """
        return self.id


    def to_json(self):
        '''
            convert cellbox to JSON

            The returned object is of the form -

                {\n
                    "geometry" (String): POLYGON(...),\n
                    "cx" (float): ...,\n
                    "cy" (float): ...,\n
                    "dcx" (float): ...,\n
                    "dcy" (float): ..., \n
                    \n
                    "agg_value_1" (float): ...,\n
                    ...,\n
                    "agg_value_n" (float): ...\n
                }\n
            Returns:
                cell_json (dict): A JSON parsable dictionary representation of this AggregatedCellBox
        '''
        cell_json = {
            "geometry": str(Polygon(self.get_bounds().get_bounds())),
            'cx': float(self.get_bounds().getcx()),
            'cy': float(self.get_bounds().getcy()),
            'dcx': float(self.get_bounds().getdcx()),
            'dcy': float(self.get_bounds().getdcy()),
           
        }

        cell_json.update(self.get_agg_data())
        cell_json['id'] = self.get_id()
        
        return cell_json



    def contains_point(self, lat, long):
        """
            Returns true if a given lat/long coordinate is contained within this cellbox.

            Args:
                lat (float): latitude of a given point
                long (float): longitude of a given point

            Returns:
                contains_points (bool): True if this CellBox contains a point given by
                    parameters (lat, long)
        """
        if (lat >= self.boundary.get_lat_min()) & (lat <= self.boundary.get_lat_max()):
            if (long >= self.boundary.get_long_min()) & (long <= self.boundary.get_long_max()):
                return True
        return False
    



from meshiphi.dataloaders.scalar.abstract_scalar import ScalarDataLoader

import xarray as xr
import numpy as np

from datetime import datetime
from os.path import basename
class ERA5WindMagDataLoader(ScalarDataLoader):
    def import_data(self, bounds):
        """
        Reads in data from an ERA5 NetCDF file.
        Renames coordinates to 'lat' and 'long'

        Args:
            bounds (Boundary): Initial boundary to limit the dataset to

        Returns:
            xr.Dataset:
                ERA5 wave dataset within limits of bounds.
                Dataset has coordinates 'lat', 'long', and variable 'wind_mag'
        """
        time_range = [datetime.strptime(time_str, "%Y-%m-%d") 
                      for time_str in bounds.get_time_range()]
        # Reduce files to those within date range
        self.files = [file for file in self.files 
                      if time_range[0] \
                      <= datetime.strptime(basename(file)[10:-3], "%Y-%m-%d") \
                      <= time_range[1]]
        # Open Dataset
        if len(self.files) == 1:    data = xr.open_dataset(self.files[0])
        else:                       data = xr.open_mfdataset(self.files).compute()
        # Change column names
        data = data.rename({'latitude': 'lat',
                            'longitude': 'long'})
        # Change domain of dataset from [0:360) to [-180:180)
        data = data.assign_coords(long=((data.long + 180) % 360) - 180)
        # Sort the 'long' axis so that sel() will work
        data = data.sortby('long')

        # Calculate magnitude from vector components
        data['wind_mag'] = np.sqrt(data['u10']**2 + data['v10']**2)

        # Limit to just rnn variables
        data = data['wind_mag'].to_dataset()
        # Reverse order of lat as array goes from max to min
        data = data.reindex(lat=data.lat[::-1])
        # Trim to initial datapoints
        data = self.trim_datapoints(bounds, data=data)
        
        return data
